# Copyright (c) 2013-2018 Nuand LLC
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


__doc__ = """
CFFI-based Python 3 binding to libbladeRF

Implements a Pythonic binding to libbladeRF, the library for interfacing
with the Nuand bladeRF product line.

See help(bladerf.BladeRF) for more information.

Basic usage:
>>> import bladerf
>>> b = bladerf.BladeRF()
"""

from bladerf import _bladerf
from bladerf import _tool

BladeRF = _bladerf.BladeRF

get_bootloader_list = _bladerf.get_bootloader_list
get_device_list = _bladerf.get_device_list
load_fw_from_bootloader = _bladerf.load_fw_from_bootloader
set_verbosity = _bladerf.set_verbosity
version = _bladerf.version

RX = _bladerf.RX
TX = _bladerf.TX
CHANNEL_RX = _bladerf.CHANNEL_RX
CHANNEL_TX = _bladerf.CHANNEL_TX

main = _tool.main

__all__ = ['BladeRF', 'get_bootloader_list', 'get_device_list',
           'load_fw_from_bootloader', 'set_verbosity', 'version', 'RX', 'TX',
           'CHANNEL_RX', 'CHANNEL_TX', 'main']



#!/usr/bin/env python3
#
# Convert a given bladeRF SC16Q11 binary samples file to CSV.
#

import csv
import struct
import sys
from   pathlib import Path

def chunked_read( fobj, chunk_bytes = 4*1024 ):
    while True:
        data = fobj.read(chunk_bytes)
        if( not data ):
            break
        else:
            yield data

def bin2csv( binfile = None, csvfile = None, chunk_bytes = 4*1024 ):
    with open(binfile, 'rb') as b:
        with open(csvfile, 'w') as c:
            csvwriter = csv.writer(c, delimiter=',')
            count = 0
            for data in chunked_read(b, chunk_bytes = chunk_bytes):
                count += len(data)
                for i in range(0, len(data), 4):
                    sig_i, = struct.unpack('<h', data[i:i+2])
                    sig_q, = struct.unpack('<h', data[i+2:i+4])
                    csvwriter.writerow( [sig_i, sig_q] )
    print( "Processed", str(count//2//2), "samples." )

if( len(sys.argv) != 3 ):
    print( "Usage:", sys.argv[0], "in.bin out.csv" )
    sys.exit(-1)
else:
    binfile = sys.argv[1]
    csvfile = sys.argv[2]
    if( Path(binfile).is_file() ):
        bin2csv( binfile, csvfile )
    else:
        print( "File does not exist:", binfile )
        sys.exit(-1)



# Copyright (c) 2013-2018 Nuand LLC
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__doc__ = """
Reads in a libbladeRF header file and produces a string ready for CFFI to
parse.

This is NOT a general-purpose tool; it should only be used on libbladeRF,
as it makes dangerous assumptions.
"""

from pycparser import parse_file, c_generator
import os
import sys
import textwrap

# This is the default for the Ubuntu python3-pycparser package
FAKE_LIBC_INCLUDE_DIR = "/usr/share/python3-pycparser/fake_libc_include"

# Lines matching these will be omitted from the output for brevity
STOPLIST = [
    "typedef struct Display Display;",
    "typedef unsigned long XID;",
    "typedef unsigned long VisualID;",
    "typedef XID Window;",
    "typedef void *MirEGLNativeWindowType;",
    "typedef void *MirEGLNativeDisplayType;",
    "typedef struct MirConnection MirConnection;",
    "typedef struct MirSurface MirSurface;",
    "typedef struct MirSurfaceSpec MirSurfaceSpec;",
    "typedef struct MirScreencast MirScreencast;",
    "typedef struct MirPromptSession MirPromptSession;",
    "typedef struct MirBufferStream MirBufferStream;",
    "typedef struct MirPersistentId MirPersistentId;",
    "typedef struct MirBlob MirBlob;",
    "typedef struct MirDisplayConfig MirDisplayConfig;",
    "typedef struct xcb_connection_t xcb_connection_t;",
    "typedef uint32_t xcb_window_t;",
    "typedef uint32_t xcb_visualid_t;",
]


class MyCGenerator(c_generator.CGenerator):
    # Clean up enumeration printouts...
    def _generate_enum_body(self, members):
        # `[:-2] + '\n'` removes the final `,` from the enumerator list
        return ''.join(self.visit(value) for value in members)[:-2] + '\n'

    def _generate_enum(self, n):
        """ Generates code for structs, unions, and enums. name should be
            'struct', 'union', or 'enum'.
        """
        members = () if n.values is None else n.values.enumerators
        s = 'enum ' + (n.name or '')
        if members:
            s += '\n'
            s += self._make_indent()
            self.indent_level += 2
            s += '{\n'
            s += self._generate_enum_body(members)
            self.indent_level -= 2
            s += self._make_indent() + '}'
        return s

    def visit_Enumerator(self, n):
        val = self.visit(n.value)
        return '{indent}{name}{equals}{value},\n'.format(
            indent=self._make_indent(),
            name=n.name,
            equals=" = " if val else "",
            value=val,
        )

    def visit_Enum(self, n):
        return self._generate_enum(n)


def parse_header(filename):
    return parse_file(filename,
                      use_cpp=True,
                      cpp_args=[
                        r'-I{}'.format(os.path.dirname(filename)),
                        r'-I{}'.format(FAKE_LIBC_INCLUDE_DIR),
                        r'-D_DOXYGEN_ONLY_'])


def ast_to_c(ast):
    gen = MyCGenerator()
    return gen.visit(ast)


def generate_cdef(data, print_omitted_lines=False):
    for line in data.split('\n'):
        # Filter typedefs included from system headers
        if 'bladerf' not in line and ('typedef int' in line or
                                      'typedef _Bool' in line or
                                      line in STOPLIST):
            if print_omitted_lines:
                line = "/* Omitted: {} */".format(line)
            else:
                line = ""

        # If we have an expression like:
        #   char magic[BLADERF_IMAGE_MAGIC_LEN + 1];
        # We need to boil it down to:
        #   char magic[8];
        if '[' in line and ']' in line:
            bstart = line.find('[')+1
            bend = line.find(']')
            prestuff = line[:bstart]
            poststuff = line[bend:]
            stuff = line[bstart:bend]
            # XXX: this is wicked dangerous to run on untrusted code
            evalstuff = str(eval(stuff))

            newline = prestuff + evalstuff + poststuff

            if newline != line:
                line = "{} /* Original: {}{}{} */".format(
                    newline, prestuff.strip(), stuff, poststuff)

        if line:
            yield line.rstrip()


def main():
    if len(sys.argv) < 2:
        print("usage: {} headerfile.h [headerfile.h...]".format(sys.argv[0]))
        sys.exit(1)

    for fn in sys.argv[1:]:
        f = parse_header(fn)
        fs = ast_to_c(f)
        print("# Generated from {} by {}".format(
            os.path.basename(fn),
            os.path.basename(sys.argv[0])))
        print()
        print("header = \"\"\"")
        for l in generate_cdef(fs):
            print('  ' + '\n    '.join(textwrap.wrap(l)))
        print("\"\"\"")


if __name__ == '__main__':
    main()



# Copyright (c) 2013-2018 Nuand LLC
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import io
import argparse

from bladerf import _bladerf


__version__ = "1.1.0"


def _bool_to_onoff(val):
    if val is None:
        return "Unsupported"
    elif val:
        return "On"
    else:
        return "Off"


def _strify_list(l):
    return ', '.join([str(x) for x in l])


def _print_channel_details(ch, verbose):
    print("  {}:".format(ch))

    if verbose:
        print("    Is TX        ", ch.is_tx)

    print("    Gain         ", ch.gain)

    if not ch.is_tx:
        print("    Gain Mode    ", ch.gain_mode)
        if verbose:
            print("        Modes    ", _strify_list(ch.gain_modes))

    if not ch.is_tx:
        print("    Symbol RSSI  ", ch.symbol_rssi)

        if verbose:
            print("        Struct   ", repr(ch.rssi))

    print("    Frequency    ", ch.frequency)
    if verbose:
        print("        Range    ", repr(ch.frequency_range))

    print("    Bandwidth    ", ch.bandwidth)
    if verbose:
        print("        Range    ", repr(ch.bandwidth_range))

    print("    Sample Rate  ", ch.sample_rate)
    if verbose:
        print("        Range    ", repr(ch.sample_rate_range))

    if verbose:
        print("    RF Port      ", ch.rf_port)
        print("        Ports    ", _strify_list(ch.rf_ports))
        print("    Bias-T Power ", _bool_to_onoff(ch.bias_tee))


def _print_cmd_info(device=None, devinfo=None, verbose=False):
    b = _bladerf.BladeRF(device, devinfo)

    devinfo = b.get_devinfo()

    flash_size, fs_guessed = b.flash_size

    if verbose:
        print("Object           ", repr(b))
    print("Board Name       ", b.board_name)
    print("Device Speed     ", b.device_speed.name)
    print("FPGA Size        ", b.fpga_size)
    print("FPGA Configured  ", b.fpga_configured)
    if b.is_fpga_configured():
        print("FPGA Version     ", b.fpga_version)
    else:
        print("FPGA Version     ", "Not Configured")
    print("Flash Size       ", (flash_size >> 17), "Mbit",
          "(assumed)" if fs_guessed else "")
    print("Firmware Version ", b.fw_version)

    if verbose:
        print("Clock Select     ", b.clock_select)
        print("Clock Output     ", _bool_to_onoff(b.clock_output))
        print("Power Source     ", b.power_source)
        print("Bus Voltage       {:.3f} V".format(b.bus_voltage or 0))
        print("Bus Current       {:.3f} A".format(b.bus_current or 0))
        print("Bus Power         {:.3f} W".format(b.bus_power or 0))
        print("RFIC temperature  {:.3f} C".format(b.rfic_temperature or 0))

        print("Loopback         ", b.loopback)
        print("    Modes        ", _strify_list(b.loopback_modes))
        print("RX Mux           ", b.rx_mux)

        print("PLL Mode         ", _bool_to_onoff(b.pll_enable))
        if b.pll_enable:
            print("PLL Refclk       ", b.pll_refclk)
            print("PLL Locked       ", _bool_to_onoff(b.pll_locked))

    print("RX Channel Count ", b.rx_channel_count)
    for c in range(b.rx_channel_count):
        ch = b.Channel(_bladerf.CHANNEL_RX(c))
        _print_channel_details(ch, verbose)

    print("TX Channel Count ", b.tx_channel_count)
    for c in range(b.tx_channel_count):
        ch = b.Channel(_bladerf.CHANNEL_TX(c))
        _print_channel_details(ch, verbose)

    b.close()


def cmd_info(device=None, verbose=False, **kwargs):
    if device is None:
        try:
            devinfos = _bladerf.get_device_list()
        except _bladerf.NoDevError:
            print("No bladeRF devices available.")
            return

        print("*** Devices found:", len(devinfos))

        for idx, dev in enumerate(devinfos):
            print()
            print("*** Device", idx)
            _print_cmd_info(devinfo=dev, verbose=verbose)

    else:
        _print_cmd_info(device=device, verbose=verbose)


def cmd_probe(**kwargs):
    try:
        devinfos = _bladerf.get_device_list()
        print("\n".join([str(devinfo) for devinfo in devinfos]))
    except _bladerf.NoDevError:
        print("No bladeRF devices available.")

    try:
        devinfos = _bladerf.get_bootloader_list()
        print("*** Detected one or more FX3-based devices in bootloader mode:")
        print("Bootloader", "\n".join([str(devinfo) for devinfo in devinfos]))
    except _bladerf.NoDevError:
        pass


def cmd_flash_fw(path, device=None, **kwargs):
    b = _bladerf.BladeRF(device)
    b.flash_firmware(path)
    b.close()


def cmd_recover_fw(path, device=None, **kwargs):
    if device is None:
        # Let's guess
        print("No device specified, trying to find one...")
        try:
            devinfos = _bladerf.get_bootloader_list()
        except _bladerf.NoDevError:
            print("No devices in bootloader recovery mode found.")
            return

        if len(devinfos) > 1:
            print("Multiple devices in bootloader recovery mode found.")
            print("Please specify one:")
            print("\n".join([str(devinfo) for devinfo in devinfos]))
            return

        device = "{backend}:device={usb_bus}:{usb_addr}".format(
                    **devinfos[0]._asdict())
        print("Choosing", device)
        print(devinfos[0])

    print("Calling load_fw_from_bootloader...")
    _bladerf.load_fw_from_bootloader(device_identifier=device, file=path)
    print("Complete")


def cmd_load_fpga(path, device=None, **kwargs):
    b = _bladerf.BladeRF(device)
    b.load_fpga(path)
    b.close()


def cmd_flash_fpga(path, device=None, **kwargs):
    b = _bladerf.BladeRF(device)
    b.flash_fpga(path)
    b.close()


def cmd_erase_fpga(device=None, **kwargs):
    b = _bladerf.BladeRF(device)
    b.erase_stored_fpga()
    b.close()


def cmd_rx(outfile, frequency, rate, gain, num_samples, device=None,
           channel=_bladerf.CHANNEL_RX(0), **kwargs):
    b = _bladerf.BladeRF(device)
    ch = b.Channel(channel)

    # Get underlying binary stream for stdout case
    if isinstance(outfile, io.TextIOWrapper):
        outfile = outfile.buffer

    # Configure BladeRF
    ch.frequency = frequency
    ch.sample_rate = rate
    ch.gain = gain

    # Setup synchronous stream
    b.sync_config(layout=_bladerf.ChannelLayout.RX_X1,
                  fmt=_bladerf.Format.SC16_Q11,
                  num_buffers=16,
                  buffer_size=8192,
                  num_transfers=8,
                  stream_timeout=3500)

    # Enable module
    ch.enable = True

    # Create buffer
    bytes_per_sample = 4
    buf = bytearray(1024*bytes_per_sample)
    num_samples_read = 0

    try:
        while True:
            if num_samples > 0 and num_samples_read == num_samples:
                break
            elif num_samples > 0:
                num = min(len(buf)//bytes_per_sample,
                          num_samples-num_samples_read)
            else:
                num = len(buf)//bytes_per_sample

            # Read into buffer
            b.sync_rx(buf, num)
            num_samples_read += num

            # Write to file
            outfile.write(buf[:num*bytes_per_sample])
    except KeyboardInterrupt:
        pass

    # Disable module
    ch.enable = False

    b.close()
    outfile.close()


def cmd_tx(infile, frequency, rate, gain, device=None,
           channel=_bladerf.CHANNEL_TX(0), **kwargs):
    b = _bladerf.BladeRF(device)
    ch = b.Channel(channel)

    # Get underlying binary stream for stdin case
    infile = infile.buffer if isinstance(infile, io.TextIOWrapper) else infile

    # Configure BladeRF
    ch.frequency = frequency
    ch.sample_rate = rate
    ch.gain = gain

    # Setup stream
    b.sync_config(layout=_bladerf.ChannelLayout.TX_X1,
                  fmt=_bladerf.Format.SC16_Q11,
                  num_buffers=16,
                  buffer_size=8192,
                  num_transfers=8,
                  stream_timeout=3500)

    # Enable module
    ch.enable = True

    # Create buffer
    bytes_per_sample = 4
    buf = bytearray(1024*bytes_per_sample)

    try:
        while True:
            # Read samples from file into buf
            num = infile.readinto(buf)
            if num == 0:
                break

            # Convert number of bytes read to samples
            num //= bytes_per_sample

            # Write to bladeRF
            b.sync_tx(buf, num)
    except KeyboardInterrupt:
        pass

    # Disable module
    ch.enable = False

    b.close()
    infile.close()


def cmd_version(**kwargs):
    print("libbladeRF version:   ", _bladerf.version())
    print("bladerf-tool version: ", "v" + __version__)


def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(title='subcommands')

    parser.add_argument('--device', '-d', help='device identifier')

    parser_info = subparsers.add_parser('info', help='get device info')
    parser_info.add_argument('--verbose', '-v', help='print more info',
                             action='store_true')
    parser_info.set_defaults(func=cmd_info)

    parser_probe = subparsers.add_parser('probe', help='probe for devices')
    parser_probe.set_defaults(func=cmd_probe)

    parser_flash_fw = subparsers.add_parser('flash_fw', help='flash firmware')
    parser_flash_fw.add_argument('path', help='firmware image path')
    parser_flash_fw.set_defaults(func=cmd_flash_fw)

    parser_recover_fw = subparsers.add_parser('recover_fw',
                                              help='recover firmware')
    parser_recover_fw.add_argument('path', help='firmware image path')
    parser_recover_fw.set_defaults(func=cmd_recover_fw)

    parser_load_fpga = subparsers.add_parser('load_fpga', help='load fpga')
    parser_load_fpga.add_argument('path', help='fpga image path')
    parser_load_fpga.set_defaults(func=cmd_load_fpga)

    parser_flash_fpga = subparsers.add_parser('flash_fpga', help='flash fpga')
    parser_flash_fpga.add_argument('path', help='fpga image path')
    parser_flash_fpga.set_defaults(func=cmd_flash_fpga)

    parser_erase_fpga = subparsers.add_parser('erase_fpga', help='erase fpga')
    parser_erase_fpga.set_defaults(func=cmd_erase_fpga)

    parser_rx = subparsers.add_parser('rx', help='receive samples',
                                      description='Receive IQ samples. '
                                                  'Sample format is FIXME.')
    parser_rx.add_argument('outfile', metavar='filename',
                           help="IQ samples filename ('-' for stdout)",
                           type=argparse.FileType('wb'))
    parser_rx.add_argument('frequency', help='frequency in Hz', type=float)
    parser_rx.add_argument('rate', help='sample rate in Hz', type=float)
    parser_rx.add_argument('--gain', '-g', help='gain in dB (default 0)',
                           type=float, default=0)
    parser_rx.add_argument('--num-samples', '-n',
                           help='number of samples (default 0 for inf)',
                           type=int, default=0)
    parser_rx.set_defaults(func=cmd_rx)

    parser_tx = subparsers.add_parser('tx', help='transmit samples',
                                      description='Transmit IQ samples. '
                                                  'Sample format is FIXME.')
    parser_tx.add_argument('infile', metavar='filename',
                           help="IQ samples filename ('-' for stdin)",
                           type=argparse.FileType('rb'))
    parser_tx.add_argument('frequency', help='frequency in Hz', type=float)
    parser_tx.add_argument('rate', help='sample rate in Hz', type=float)
    parser_tx.add_argument('--gain', '-g', help='gain in dB (default 0)',
                           type=float, default=0)
    parser_tx.set_defaults(func=cmd_tx)

    parser_version = subparsers.add_parser('version',
                                           help='print version info')
    parser_version.set_defaults(func=cmd_version)

    args = parser.parse_args()
    if 'func' not in args:
        parser.print_help()
        parser.exit()

    args.func(**vars(args))


if __name__ == "__main__":
    main()



# Generated from libbladeRF.h by import_header.py

header = """
  struct bladerf;
  typedef enum
  {
    BLADERF_BACKEND_ANY,
    BLADERF_BACKEND_LINUX,
    BLADERF_BACKEND_LIBUSB,
    BLADERF_BACKEND_CYPRESS,
    BLADERF_BACKEND_DUMMY = 100
  } bladerf_backend;
  struct bladerf_devinfo
  {
    bladerf_backend backend;
    char serial[33];
    uint8_t usb_bus;
    uint8_t usb_addr;
    unsigned int instance;
    char manufacturer[33];
    char product[33];
  };
  struct bladerf_backendinfo
  {
    int handle_count;
    void *handle;
    int lock_count;
    void *lock;
  };
  int bladerf_open(struct bladerf **device, const char
    *device_identifier);
  void bladerf_close(struct bladerf *device);
  int bladerf_open_with_devinfo(struct bladerf **device, struct
    bladerf_devinfo *devinfo);
  int bladerf_get_device_list(struct bladerf_devinfo **devices);
  void bladerf_free_device_list(struct bladerf_devinfo *devices);
  void bladerf_init_devinfo(struct bladerf_devinfo *info);
  int bladerf_get_devinfo(struct bladerf *dev, struct bladerf_devinfo
    *info);
  int bladerf_get_backendinfo(struct bladerf *dev, struct
    bladerf_backendinfo *info);
  int bladerf_get_devinfo_from_str(const char *devstr, struct
    bladerf_devinfo *info);
  bool bladerf_devinfo_matches(const struct bladerf_devinfo *a, const
    struct bladerf_devinfo *b);
  bool bladerf_devstr_matches(const char *dev_str, struct
    bladerf_devinfo *info);
  const char *bladerf_backend_str(bladerf_backend backend);
  void bladerf_set_usb_reset_on_open(bool enabled);
  struct bladerf_range
  {
    int64_t min;
    int64_t max;
    int64_t step;
    float scale;
  };
  struct bladerf_serial
  {
    char serial[33];
  };
  struct bladerf_version
  {
    uint16_t major;
    uint16_t minor;
    uint16_t patch;
    const char *describe;
  };
  typedef enum
  {
    BLADERF_FPGA_UNKNOWN = 0,
    BLADERF_FPGA_40KLE = 40,
    BLADERF_FPGA_115KLE = 115,
    BLADERF_FPGA_A4 = 49,
    BLADERF_FPGA_A5 = 77,
    BLADERF_FPGA_A9 = 301
  } bladerf_fpga_size;
  typedef enum
  {
    BLADERF_DEVICE_SPEED_UNKNOWN,
    BLADERF_DEVICE_SPEED_HIGH,
    BLADERF_DEVICE_SPEED_SUPER
  } bladerf_dev_speed;
  typedef enum
  {
    BLADERF_FPGA_SOURCE_UNKNOWN = 0,
    BLADERF_FPGA_SOURCE_FLASH = 1,
    BLADERF_FPGA_SOURCE_HOST = 2
  } bladerf_fpga_source;
  int bladerf_get_serial(struct bladerf *dev, char *serial);
  int bladerf_get_serial_struct(struct bladerf *dev, struct
    bladerf_serial *serial);
  int bladerf_get_fpga_size(struct bladerf *dev, bladerf_fpga_size
    *size);
  int bladerf_get_fpga_bytes(struct bladerf *dev, size_t *size);
  int bladerf_get_flash_size(struct bladerf *dev, uint32_t *size, bool
    *is_guess);
  int bladerf_fw_version(struct bladerf *dev, struct bladerf_version
    *version);
  int bladerf_is_fpga_configured(struct bladerf *dev);
  int bladerf_fpga_version(struct bladerf *dev, struct bladerf_version
    *version);
  int bladerf_get_fpga_source(struct bladerf *dev, bladerf_fpga_source
    *source);
  bladerf_dev_speed bladerf_device_speed(struct bladerf *dev);
  const char *bladerf_get_board_name(struct bladerf *dev);
  typedef int bladerf_channel;
  typedef bladerf_channel bladerf_module;
  typedef enum
  {
    BLADERF_RX = 0,
    BLADERF_TX = 1
  } bladerf_direction;
  typedef enum
  {
    BLADERF_RX_X1 = 0,
    BLADERF_TX_X1 = 1,
    BLADERF_RX_X2 = 2,
    BLADERF_TX_X2 = 3
  } bladerf_channel_layout;
  size_t bladerf_get_channel_count(struct bladerf *dev,
    bladerf_direction dir);
  typedef int bladerf_gain;
  typedef enum
  {
    BLADERF_GAIN_DEFAULT,
    BLADERF_GAIN_MGC,
    BLADERF_GAIN_FASTATTACK_AGC,
    BLADERF_GAIN_SLOWATTACK_AGC,
    BLADERF_GAIN_HYBRID_AGC
  } bladerf_gain_mode;
  struct bladerf_gain_modes
  {
    const char *name;
    bladerf_gain_mode mode;
  };
  int bladerf_set_gain(struct bladerf *dev, bladerf_channel ch,
    bladerf_gain gain);
  int bladerf_get_gain(struct bladerf *dev, bladerf_channel ch,
    bladerf_gain *gain);
  int bladerf_set_gain_mode(struct bladerf *dev, bladerf_channel ch,
    bladerf_gain_mode mode);
  int bladerf_get_gain_mode(struct bladerf *dev, bladerf_channel ch,
    bladerf_gain_mode *mode);
  int bladerf_get_gain_modes(struct bladerf *dev, bladerf_channel ch,
    const struct bladerf_gain_modes **modes);
  int bladerf_get_gain_range(struct bladerf *dev, bladerf_channel ch,
    const struct bladerf_range **range);
  int bladerf_set_gain_stage(struct bladerf *dev, bladerf_channel ch,
    const char *stage, bladerf_gain gain);
  int bladerf_get_gain_stage(struct bladerf *dev, bladerf_channel ch,
    const char *stage, bladerf_gain *gain);
  int bladerf_get_gain_stage_range(struct bladerf *dev, bladerf_channel
    ch, const char *stage, const struct bladerf_range **range);
  int bladerf_get_gain_stages(struct bladerf *dev, bladerf_channel ch,
    const char **stages, size_t count);
  typedef unsigned int bladerf_sample_rate;
  struct bladerf_rational_rate
  {
    uint64_t integer;
    uint64_t num;
    uint64_t den;
  };
  int bladerf_set_sample_rate(struct bladerf *dev, bladerf_channel ch,
    bladerf_sample_rate rate, bladerf_sample_rate *actual);
  int bladerf_set_rational_sample_rate(struct bladerf *dev,
    bladerf_channel ch, struct bladerf_rational_rate *rate, struct
    bladerf_rational_rate *actual);
  int bladerf_get_sample_rate(struct bladerf *dev, bladerf_channel ch,
    bladerf_sample_rate *rate);
  int bladerf_get_sample_rate_range(struct bladerf *dev, bladerf_channel
    ch, const struct bladerf_range **range);
  int bladerf_get_rational_sample_rate(struct bladerf *dev,
    bladerf_channel ch, struct bladerf_rational_rate *rate);
  typedef unsigned int bladerf_bandwidth;
  int bladerf_set_bandwidth(struct bladerf *dev, bladerf_channel ch,
    bladerf_bandwidth bandwidth, bladerf_bandwidth *actual);
  int bladerf_get_bandwidth(struct bladerf *dev, bladerf_channel ch,
    bladerf_bandwidth *bandwidth);
  int bladerf_get_bandwidth_range(struct bladerf *dev, bladerf_channel
    ch, const struct bladerf_range **range);
  typedef uint64_t bladerf_frequency;
  int bladerf_select_band(struct bladerf *dev, bladerf_channel ch,
    bladerf_frequency frequency);
  int bladerf_set_frequency(struct bladerf *dev, bladerf_channel ch,
    bladerf_frequency frequency);
  int bladerf_get_frequency(struct bladerf *dev, bladerf_channel ch,
    bladerf_frequency *frequency);
  int bladerf_get_frequency_range(struct bladerf *dev, bladerf_channel
    ch, const struct bladerf_range **range);
  typedef enum
  {
    BLADERF_LB_NONE = 0,
    BLADERF_LB_FIRMWARE,
    BLADERF_LB_BB_TXLPF_RXVGA2,
    BLADERF_LB_BB_TXVGA1_RXVGA2,
    BLADERF_LB_BB_TXLPF_RXLPF,
    BLADERF_LB_BB_TXVGA1_RXLPF,
    BLADERF_LB_RF_LNA1,
    BLADERF_LB_RF_LNA2,
    BLADERF_LB_RF_LNA3,
    BLADERF_LB_RFIC_BIST
  } bladerf_loopback;
  struct bladerf_loopback_modes
  {
    const char *name;
    bladerf_loopback mode;
  };
  int bladerf_get_loopback_modes(struct bladerf *dev, const struct
    bladerf_loopback_modes **modes);
  bool bladerf_is_loopback_mode_supported(struct bladerf *dev,
    bladerf_loopback mode);
  int bladerf_set_loopback(struct bladerf *dev, bladerf_loopback lb);
  int bladerf_get_loopback(struct bladerf *dev, bladerf_loopback *lb);
  typedef enum
  {
    BLADERF_TRIGGER_ROLE_INVALID = -1,
    BLADERF_TRIGGER_ROLE_DISABLED,
    BLADERF_TRIGGER_ROLE_MASTER,
    BLADERF_TRIGGER_ROLE_SLAVE
  } bladerf_trigger_role;
  typedef enum
  {
    BLADERF_TRIGGER_INVALID = -1,
    BLADERF_TRIGGER_J71_4,
    BLADERF_TRIGGER_J51_1,
    BLADERF_TRIGGER_MINI_EXP_1,
    BLADERF_TRIGGER_USER_0 = 128,
    BLADERF_TRIGGER_USER_1,
    BLADERF_TRIGGER_USER_2,
    BLADERF_TRIGGER_USER_3,
    BLADERF_TRIGGER_USER_4,
    BLADERF_TRIGGER_USER_5,
    BLADERF_TRIGGER_USER_6,
    BLADERF_TRIGGER_USER_7
  } bladerf_trigger_signal;
  struct bladerf_trigger
  {
    bladerf_channel channel;
    bladerf_trigger_role role;
    bladerf_trigger_signal signal;
    uint64_t options;
  };
  int bladerf_trigger_init(struct bladerf *dev, bladerf_channel ch,
    bladerf_trigger_signal signal, struct bladerf_trigger *trigger);
  int bladerf_trigger_arm(struct bladerf *dev, const struct
    bladerf_trigger *trigger, bool arm, uint64_t resv1, uint64_t resv2);
  int bladerf_trigger_fire(struct bladerf *dev, const struct
    bladerf_trigger *trigger);
  int bladerf_trigger_state(struct bladerf *dev, const struct
    bladerf_trigger *trigger, bool *is_armed, bool *has_fired, bool
    *fire_requested, uint64_t *resv1, uint64_t *resv2);
  typedef enum
  {
    BLADERF_RX_MUX_INVALID = -1,
    BLADERF_RX_MUX_BASEBAND = 0x0,
    BLADERF_RX_MUX_12BIT_COUNTER = 0x1,
    BLADERF_RX_MUX_32BIT_COUNTER = 0x2,
    BLADERF_RX_MUX_DIGITAL_LOOPBACK = 0x4
  } bladerf_rx_mux;
  int bladerf_set_rx_mux(struct bladerf *dev, bladerf_rx_mux mux);
  int bladerf_get_rx_mux(struct bladerf *dev, bladerf_rx_mux *mode);
  typedef uint64_t bladerf_timestamp;
  struct bladerf_quick_tune
  {
    union
    {
      struct
      {
        uint8_t freqsel;
        uint8_t vcocap;
        uint16_t nint;
        uint32_t nfrac;
        uint8_t flags;
        uint8_t xb_gpio;
      };
      struct
      {
        uint16_t nios_profile;
        uint8_t rffe_profile;
        uint8_t port;
        uint8_t spdt;
      };
    };
  };
  int bladerf_schedule_retune(struct bladerf *dev, bladerf_channel ch,
    bladerf_timestamp timestamp, bladerf_frequency frequency, struct
    bladerf_quick_tune *quick_tune);
  int bladerf_cancel_scheduled_retunes(struct bladerf *dev,
    bladerf_channel ch);
  int bladerf_get_quick_tune(struct bladerf *dev, bladerf_channel ch,
    struct bladerf_quick_tune *quick_tune);
  typedef int16_t bladerf_correction_value;
  typedef enum
  {
    BLADERF_CORR_DCOFF_I,
    BLADERF_CORR_DCOFF_Q,
    BLADERF_CORR_PHASE,
    BLADERF_CORR_GAIN
  } bladerf_correction;
  int bladerf_set_correction(struct bladerf *dev, bladerf_channel ch,
    bladerf_correction corr, bladerf_correction_value value);
  int bladerf_get_correction(struct bladerf *dev, bladerf_channel ch,
    bladerf_correction corr, bladerf_correction_value *value);
  typedef enum
  {
    BLADERF_FORMAT_SC16_Q11,
    BLADERF_FORMAT_SC16_Q11_META,
    BLADERF_FORMAT_PACKET_META,
    BLADERF_FORMAT_SC8_Q7,
    BLADERF_FORMAT_SC8_Q7_META
  } bladerf_format;
  struct bladerf_metadata
  {
    bladerf_timestamp timestamp;
    uint32_t flags;
    uint32_t status;
    unsigned int actual_count;
    uint8_t reserved[32];
  };
  int bladerf_interleave_stream_buffer(bladerf_channel_layout layout,
    bladerf_format format, unsigned int buffer_size, void *samples);
  int bladerf_deinterleave_stream_buffer(bladerf_channel_layout layout,
    bladerf_format format, unsigned int buffer_size, void *samples);
  int bladerf_enable_module(struct bladerf *dev, bladerf_channel ch,
    bool enable);
  int bladerf_get_timestamp(struct bladerf *dev, bladerf_direction dir,
    bladerf_timestamp *timestamp);
  int bladerf_sync_config(struct bladerf *dev, bladerf_channel_layout
    layout, bladerf_format format, unsigned int num_buffers, unsigned int
    buffer_size, unsigned int num_transfers, unsigned int stream_timeout);
  int bladerf_sync_tx(struct bladerf *dev, const void *samples, unsigned
    int num_samples, struct bladerf_metadata *metadata, unsigned int
    timeout_ms);
  int bladerf_sync_rx(struct bladerf *dev, void *samples, unsigned int
    num_samples, struct bladerf_metadata *metadata, unsigned int
    timeout_ms);
  struct bladerf_stream;
  typedef void *(*bladerf_stream_cb)(struct bladerf *dev, struct
    bladerf_stream *stream, struct bladerf_metadata *meta, void *samples,
    size_t num_samples, void *user_data);
  int bladerf_init_stream(struct bladerf_stream **stream, struct bladerf
    *dev, bladerf_stream_cb callback, void ***buffers, size_t num_buffers,
    bladerf_format format, size_t samples_per_buffer, size_t
    num_transfers, void *user_data);
  int bladerf_stream(struct bladerf_stream *stream,
    bladerf_channel_layout layout);
  int bladerf_submit_stream_buffer(struct bladerf_stream *stream, void
    *buffer, unsigned int timeout_ms);
  int bladerf_submit_stream_buffer_nb(struct bladerf_stream *stream,
    void *buffer);
  void bladerf_deinit_stream(struct bladerf_stream *stream);
  int bladerf_set_stream_timeout(struct bladerf *dev, bladerf_direction
    dir, unsigned int timeout);
  int bladerf_get_stream_timeout(struct bladerf *dev, bladerf_direction
    dir, unsigned int *timeout);
  int bladerf_flash_firmware(struct bladerf *dev, const char *firmware);
  int bladerf_load_fpga(struct bladerf *dev, const char *fpga);
  int bladerf_flash_fpga(struct bladerf *dev, const char *fpga_image);
  int bladerf_erase_stored_fpga(struct bladerf *dev);
  int bladerf_device_reset(struct bladerf *dev);
  int bladerf_get_fw_log(struct bladerf *dev, const char *filename);
  int bladerf_jump_to_bootloader(struct bladerf *dev);
  int bladerf_get_bootloader_list(struct bladerf_devinfo **list);
  int bladerf_load_fw_from_bootloader(const char *device_identifier,
    bladerf_backend backend, uint8_t bus, uint8_t addr, const char *file);
  typedef enum
  {
    BLADERF_IMAGE_TYPE_INVALID = -1,
    BLADERF_IMAGE_TYPE_RAW,
    BLADERF_IMAGE_TYPE_FIRMWARE,
    BLADERF_IMAGE_TYPE_FPGA_40KLE,
    BLADERF_IMAGE_TYPE_FPGA_115KLE,
    BLADERF_IMAGE_TYPE_FPGA_A4,
    BLADERF_IMAGE_TYPE_FPGA_A9,
    BLADERF_IMAGE_TYPE_CALIBRATION,
    BLADERF_IMAGE_TYPE_RX_DC_CAL,
    BLADERF_IMAGE_TYPE_TX_DC_CAL,
    BLADERF_IMAGE_TYPE_RX_IQ_CAL,
    BLADERF_IMAGE_TYPE_TX_IQ_CAL,
    BLADERF_IMAGE_TYPE_FPGA_A5
  } bladerf_image_type;
  struct bladerf_image
  {
    char magic[8]; /* Original: char magic[7 + 1]; */
    uint8_t checksum[32];
    struct bladerf_version version;
    uint64_t timestamp;
    char serial[34]; /* Original: char serial[33 + 1]; */
    char reserved[128];
    bladerf_image_type type;
    uint32_t address;
    uint32_t length;
    uint8_t *data;
  };
  struct bladerf_image *bladerf_alloc_image(struct bladerf *dev,
    bladerf_image_type type, uint32_t address, uint32_t length);
  struct bladerf_image *bladerf_alloc_cal_image(struct bladerf *dev,
    bladerf_fpga_size fpga_size, uint16_t vctcxo_trim);
  void bladerf_free_image(struct bladerf_image *image);
  int bladerf_image_write(struct bladerf *dev, struct bladerf_image
    *image, const char *file);
  int bladerf_image_read(struct bladerf_image *image, const char *file);
  typedef enum
  {
    BLADERF_VCTCXO_TAMER_INVALID = -1,
    BLADERF_VCTCXO_TAMER_DISABLED = 0,
    BLADERF_VCTCXO_TAMER_1_PPS = 1,
    BLADERF_VCTCXO_TAMER_10_MHZ = 2
  } bladerf_vctcxo_tamer_mode;
  int bladerf_set_vctcxo_tamer_mode(struct bladerf *dev,
    bladerf_vctcxo_tamer_mode mode);
  int bladerf_get_vctcxo_tamer_mode(struct bladerf *dev,
    bladerf_vctcxo_tamer_mode *mode);
  int bladerf_get_vctcxo_trim(struct bladerf *dev, uint16_t *trim);
  int bladerf_trim_dac_write(struct bladerf *dev, uint16_t val);
  int bladerf_trim_dac_read(struct bladerf *dev, uint16_t *val);
  typedef enum
  {
    BLADERF_TUNING_MODE_INVALID = -1,
    BLADERF_TUNING_MODE_HOST,
    BLADERF_TUNING_MODE_FPGA
  } bladerf_tuning_mode;
  int bladerf_set_tuning_mode(struct bladerf *dev, bladerf_tuning_mode
    mode);
  int bladerf_get_tuning_mode(struct bladerf *dev, bladerf_tuning_mode
    *mode);
  int bladerf_read_trigger(struct bladerf *dev, bladerf_channel ch,
    bladerf_trigger_signal signal, uint8_t *val);
  int bladerf_write_trigger(struct bladerf *dev, bladerf_channel ch,
    bladerf_trigger_signal signal, uint8_t val);
  int bladerf_wishbone_master_read(struct bladerf *dev, uint32_t addr,
    uint32_t *data);
  int bladerf_wishbone_master_write(struct bladerf *dev, uint32_t addr,
    uint32_t val);
  int bladerf_config_gpio_read(struct bladerf *dev, uint32_t *val);
  int bladerf_config_gpio_write(struct bladerf *dev, uint32_t val);
  int bladerf_erase_flash(struct bladerf *dev, uint32_t erase_block,
    uint32_t count);
  int bladerf_erase_flash_bytes(struct bladerf *dev, uint32_t address,
    uint32_t length);
  int bladerf_read_flash(struct bladerf *dev, uint8_t *buf, uint32_t
    page, uint32_t count);
  int bladerf_read_flash_bytes(struct bladerf *dev, uint8_t *buf,
    uint32_t address, uint32_t bytes);
  int bladerf_write_flash(struct bladerf *dev, const uint8_t *buf,
    uint32_t page, uint32_t count);
  int bladerf_write_flash_bytes(struct bladerf *dev, const uint8_t *buf,
    uint32_t address, uint32_t length);
  int bladerf_lock_otp(struct bladerf *dev);
  int bladerf_read_otp(struct bladerf *dev, uint8_t *buf);
  int bladerf_write_otp(struct bladerf *dev, uint8_t *buf);
  int bladerf_set_rf_port(struct bladerf *dev, bladerf_channel ch, const
    char *port);
  int bladerf_get_rf_port(struct bladerf *dev, bladerf_channel ch, const
    char **port);
  int bladerf_get_rf_ports(struct bladerf *dev, bladerf_channel ch,
    const char **ports, unsigned int count);
  typedef enum
  {
    BLADERF_FEATURE_DEFAULT = 0,
    BLADERF_FEATURE_OVERSAMPLE
  } bladerf_feature;
  int bladerf_enable_feature(struct bladerf *dev, bladerf_feature
    feature, bool enable);
  int bladerf_get_feature(struct bladerf *dev, bladerf_feature
    *feature);
  typedef enum
  {
    BLADERF_XB_NONE = 0,
    BLADERF_XB_100,
    BLADERF_XB_200,
    BLADERF_XB_300
  } bladerf_xb;
  int bladerf_expansion_attach(struct bladerf *dev, bladerf_xb xb);
  int bladerf_expansion_get_attached(struct bladerf *dev, bladerf_xb
    *xb);
  typedef enum
  {
    BLADERF_LOG_LEVEL_VERBOSE,
    BLADERF_LOG_LEVEL_DEBUG,
    BLADERF_LOG_LEVEL_INFO,
    BLADERF_LOG_LEVEL_WARNING,
    BLADERF_LOG_LEVEL_ERROR,
    BLADERF_LOG_LEVEL_CRITICAL,
    BLADERF_LOG_LEVEL_SILENT
  } bladerf_log_level;
  void bladerf_log_set_verbosity(bladerf_log_level level);
  void bladerf_version(struct bladerf_version *version);
  const char *bladerf_strerror(int error);
  typedef enum
  {
    BLADERF_LNA_GAIN_UNKNOWN,
    BLADERF_LNA_GAIN_BYPASS,
    BLADERF_LNA_GAIN_MID,
    BLADERF_LNA_GAIN_MAX
  } bladerf_lna_gain;
  int bladerf_set_txvga2(struct bladerf *dev, int gain);
  int bladerf_get_txvga2(struct bladerf *dev, int *gain);
  int bladerf_set_txvga1(struct bladerf *dev, int gain);
  int bladerf_get_txvga1(struct bladerf *dev, int *gain);
  int bladerf_set_lna_gain(struct bladerf *dev, bladerf_lna_gain gain);
  int bladerf_get_lna_gain(struct bladerf *dev, bladerf_lna_gain *gain);
  int bladerf_set_rxvga1(struct bladerf *dev, int gain);
  int bladerf_get_rxvga1(struct bladerf *dev, int *gain);
  int bladerf_set_rxvga2(struct bladerf *dev, int gain);
  int bladerf_get_rxvga2(struct bladerf *dev, int *gain);
  typedef enum
  {
    BLADERF_SAMPLING_UNKNOWN,
    BLADERF_SAMPLING_INTERNAL,
    BLADERF_SAMPLING_EXTERNAL
  } bladerf_sampling;
  int bladerf_set_sampling(struct bladerf *dev, bladerf_sampling
    sampling);
  int bladerf_get_sampling(struct bladerf *dev, bladerf_sampling
    *sampling);
  typedef enum
  {
    BLADERF_LPF_NORMAL,
    BLADERF_LPF_BYPASSED,
    BLADERF_LPF_DISABLED
  } bladerf_lpf_mode;
  int bladerf_set_lpf_mode(struct bladerf *dev, bladerf_channel ch,
    bladerf_lpf_mode mode);
  int bladerf_get_lpf_mode(struct bladerf *dev, bladerf_channel ch,
    bladerf_lpf_mode *mode);
  typedef enum
  {
    BLADERF_SMB_MODE_INVALID = -1,
    BLADERF_SMB_MODE_DISABLED,
    BLADERF_SMB_MODE_OUTPUT,
    BLADERF_SMB_MODE_INPUT,
    BLADERF_SMB_MODE_UNAVAILBLE
  } bladerf_smb_mode;
  int bladerf_set_smb_mode(struct bladerf *dev, bladerf_smb_mode mode);
  int bladerf_get_smb_mode(struct bladerf *dev, bladerf_smb_mode *mode);
  int bladerf_set_rational_smb_frequency(struct bladerf *dev, struct
    bladerf_rational_rate *rate, struct bladerf_rational_rate *actual);
  int bladerf_set_smb_frequency(struct bladerf *dev, uint32_t rate,
    uint32_t *actual);
  int bladerf_get_rational_smb_frequency(struct bladerf *dev, struct
    bladerf_rational_rate *rate);
  int bladerf_get_smb_frequency(struct bladerf *dev, unsigned int
    *rate);
  int bladerf_expansion_gpio_read(struct bladerf *dev, uint32_t *val);
  int bladerf_expansion_gpio_write(struct bladerf *dev, uint32_t val);
  int bladerf_expansion_gpio_masked_write(struct bladerf *dev, uint32_t
    mask, uint32_t value);
  int bladerf_expansion_gpio_dir_read(struct bladerf *dev, uint32_t
    *outputs);
  int bladerf_expansion_gpio_dir_write(struct bladerf *dev, uint32_t
    outputs);
  int bladerf_expansion_gpio_dir_masked_write(struct bladerf *dev,
    uint32_t mask, uint32_t outputs);
  typedef enum
  {
    BLADERF_XB200_50M = 0,
    BLADERF_XB200_144M,
    BLADERF_XB200_222M,
    BLADERF_XB200_CUSTOM,
    BLADERF_XB200_AUTO_1DB,
    BLADERF_XB200_AUTO_3DB
  } bladerf_xb200_filter;
  typedef enum
  {
    BLADERF_XB200_BYPASS = 0,
    BLADERF_XB200_MIX
  } bladerf_xb200_path;
  typedef enum
  {
    BLADERF_XB300_TRX_INVAL = -1,
    BLADERF_XB300_TRX_TX = 0,
    BLADERF_XB300_TRX_RX,
    BLADERF_XB300_TRX_UNSET
  } bladerf_xb300_trx;
  typedef enum
  {
    BLADERF_XB300_AMP_INVAL = -1,
    BLADERF_XB300_AMP_PA = 0,
    BLADERF_XB300_AMP_LNA,
    BLADERF_XB300_AMP_PA_AUX
  } bladerf_xb300_amplifier;
  int bladerf_xb200_set_filterbank(struct bladerf *dev, bladerf_channel
    ch, bladerf_xb200_filter filter);
  int bladerf_xb200_get_filterbank(struct bladerf *dev, bladerf_channel
    ch, bladerf_xb200_filter *filter);
  int bladerf_xb200_set_path(struct bladerf *dev, bladerf_channel ch,
    bladerf_xb200_path path);
  int bladerf_xb200_get_path(struct bladerf *dev, bladerf_channel ch,
    bladerf_xb200_path *path);
  int bladerf_xb300_set_trx(struct bladerf *dev, bladerf_xb300_trx trx);
  int bladerf_xb300_get_trx(struct bladerf *dev, bladerf_xb300_trx
    *trx);
  int bladerf_xb300_set_amplifier_enable(struct bladerf *dev,
    bladerf_xb300_amplifier amp, bool enable);
  int bladerf_xb300_get_amplifier_enable(struct bladerf *dev,
    bladerf_xb300_amplifier amp, bool *enable);
  int bladerf_xb300_get_output_power(struct bladerf *dev, float *val);
  typedef enum
  {
    BLADERF_DC_CAL_INVALID = -1,
    BLADERF_DC_CAL_LPF_TUNING,
    BLADERF_DC_CAL_TX_LPF,
    BLADERF_DC_CAL_RX_LPF,
    BLADERF_DC_CAL_RXVGA2
  } bladerf_cal_module;
  int bladerf_calibrate_dc(struct bladerf *dev, bladerf_cal_module
    module);
  int bladerf_dac_write(struct bladerf *dev, uint16_t val);
  int bladerf_dac_read(struct bladerf *dev, uint16_t *val);
  int bladerf_si5338_read(struct bladerf *dev, uint8_t address, uint8_t
    *val);
  int bladerf_si5338_write(struct bladerf *dev, uint8_t address, uint8_t
    val);
  int bladerf_lms_read(struct bladerf *dev, uint8_t address, uint8_t
    *val);
  int bladerf_lms_write(struct bladerf *dev, uint8_t address, uint8_t
    val);
  struct bladerf_lms_dc_cals
  {
    int16_t lpf_tuning;
    int16_t tx_lpf_i;
    int16_t tx_lpf_q;
    int16_t rx_lpf_i;
    int16_t rx_lpf_q;
    int16_t dc_ref;
    int16_t rxvga2a_i;
    int16_t rxvga2a_q;
    int16_t rxvga2b_i;
    int16_t rxvga2b_q;
  };
  int bladerf_lms_set_dc_cals(struct bladerf *dev, const struct
    bladerf_lms_dc_cals *dc_cals);
  int bladerf_lms_get_dc_cals(struct bladerf *dev, struct
    bladerf_lms_dc_cals *dc_cals);
  int bladerf_xb_spi_write(struct bladerf *dev, uint32_t val);
  int bladerf_get_bias_tee(struct bladerf *dev, bladerf_channel ch, bool
    *enable);
  int bladerf_set_bias_tee(struct bladerf *dev, bladerf_channel ch, bool
    enable);
  int bladerf_get_rfic_register(struct bladerf *dev, uint16_t address,
    uint8_t *val);
  int bladerf_set_rfic_register(struct bladerf *dev, uint16_t address,
    uint8_t val);
  int bladerf_get_rfic_temperature(struct bladerf *dev, float *val);
  int bladerf_get_rfic_rssi(struct bladerf *dev, bladerf_channel ch,
    int32_t *pre_rssi, int32_t *sym_rssi);
  int bladerf_get_rfic_ctrl_out(struct bladerf *dev, uint8_t *ctrl_out);
  typedef enum
  {
    BLADERF_RFIC_RXFIR_BYPASS = 0,
    BLADERF_RFIC_RXFIR_CUSTOM,
    BLADERF_RFIC_RXFIR_DEC1,
    BLADERF_RFIC_RXFIR_DEC2,
    BLADERF_RFIC_RXFIR_DEC4
  } bladerf_rfic_rxfir;
  typedef enum
  {
    BLADERF_RFIC_TXFIR_BYPASS = 0,
    BLADERF_RFIC_TXFIR_CUSTOM,
    BLADERF_RFIC_TXFIR_INT1,
    BLADERF_RFIC_TXFIR_INT2,
    BLADERF_RFIC_TXFIR_INT4
  } bladerf_rfic_txfir;
  int bladerf_get_rfic_rx_fir(struct bladerf *dev, bladerf_rfic_rxfir
    *rxfir);
  int bladerf_set_rfic_rx_fir(struct bladerf *dev, bladerf_rfic_rxfir
    rxfir);
  int bladerf_get_rfic_tx_fir(struct bladerf *dev, bladerf_rfic_txfir
    *txfir);
  int bladerf_set_rfic_tx_fir(struct bladerf *dev, bladerf_rfic_txfir
    txfir);
  int bladerf_get_pll_lock_state(struct bladerf *dev, bool *locked);
  int bladerf_get_pll_enable(struct bladerf *dev, bool *enabled);
  int bladerf_set_pll_enable(struct bladerf *dev, bool enable);
  int bladerf_get_pll_refclk_range(struct bladerf *dev, const struct
    bladerf_range **range);
  int bladerf_get_pll_refclk(struct bladerf *dev, uint64_t *frequency);
  int bladerf_set_pll_refclk(struct bladerf *dev, uint64_t frequency);
  int bladerf_get_pll_register(struct bladerf *dev, uint8_t address,
    uint32_t *val);
  int bladerf_set_pll_register(struct bladerf *dev, uint8_t address,
    uint32_t val);
  typedef enum
  {
    BLADERF_UNKNOWN,
    BLADERF_PS_DC,
    BLADERF_PS_USB_VBUS
  } bladerf_power_sources;
  int bladerf_get_power_source(struct bladerf *dev,
    bladerf_power_sources *val);
  typedef enum
  {
    CLOCK_SELECT_ONBOARD,
    CLOCK_SELECT_EXTERNAL
  } bladerf_clock_select;
  int bladerf_get_clock_select(struct bladerf *dev, bladerf_clock_select
    *sel);
  int bladerf_set_clock_select(struct bladerf *dev, bladerf_clock_select
    sel);
  int bladerf_get_clock_output(struct bladerf *dev, bool *state);
  int bladerf_set_clock_output(struct bladerf *dev, bool enable);
  typedef enum
  {
    BLADERF_PMIC_CONFIGURATION,
    BLADERF_PMIC_VOLTAGE_SHUNT,
    BLADERF_PMIC_VOLTAGE_BUS,
    BLADERF_PMIC_POWER,
    BLADERF_PMIC_CURRENT,
    BLADERF_PMIC_CALIBRATION
  } bladerf_pmic_register;
  int bladerf_get_pmic_register(struct bladerf *dev,
    bladerf_pmic_register reg, void *val);
  typedef struct
  {
    uint32_t tx1_rfic_port;
    uint32_t tx1_spdt_port;
    uint32_t tx2_rfic_port;
    uint32_t tx2_spdt_port;
    uint32_t rx1_rfic_port;
    uint32_t rx1_spdt_port;
    uint32_t rx2_rfic_port;
    uint32_t rx2_spdt_port;
  } bladerf_rf_switch_config;
  int bladerf_get_rf_switch_config(struct bladerf *dev,
    bladerf_rf_switch_config *config);
"""



# Copyright (c) 2013-2018 Nuand LLC
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import enum
import collections

import cffi

from sys import platform

from ._cdef import header

ffi = cffi.FFI()

ffi.cdef(header)

if platform == "win32":
      libbladeRF = ffi.dlopen("bladerf.dll")
elif platform == "darwin":
    libbladeRF = ffi.dlopen("libbladeRF.dylib")
else:
    libbladeRF = ffi.dlopen("libbladeRF.so")


###############################################################################


# Python class wrappers for various structures

class ReturnCode(enum.Enum):
    BladeRFError = 0
    UnexpectedError = -1
    RangeError = -2
    InvalError = -3
    MemError = -4
    IOError = -5
    TimeoutError = -6
    NoDevError = -7
    UnsupportedError = -8
    MisalignedError = -9
    ChecksumError = -10
    NoFileError = -11
    UpdateFPGAError = -12
    UpdateFWError = -13
    TimePastError = -14
    QueueFullError = -15
    FPGAOpError = -16
    PermissionError = -17
    WouldBlockError = -18
    NotInitError = -19

    def __str__(self):
        return self.name


class Version(collections.namedtuple("Version", [
                "major", "minor", "patch", "describe"])):
    @staticmethod
    def from_struct(version):
        return Version(version.major, version.minor, version.patch,
                       ffi.string(version.describe).decode())

    def __str__(self):
        return "v{}.{}.{} (\"{}\")".format(*self)


class RationalRate(collections.namedtuple("RationalRate", [
                    "integer", "num", "den"])):
    @staticmethod
    def from_struct(rate):
        return RationalRate(rate.integer, rate.num, rate.den)

    def to_struct(self):
        return ffi.new("struct bladerf_rational_rate *", [
                        self.integer, self.num, self.den])

    struct = property(to_struct)


class Backend(enum.Enum):
    Any = libbladeRF.BLADERF_BACKEND_ANY
    Linux = libbladeRF.BLADERF_BACKEND_LINUX
    LibUSB = libbladeRF.BLADERF_BACKEND_LIBUSB
    Cypress = libbladeRF.BLADERF_BACKEND_CYPRESS
    Dummy = libbladeRF.BLADERF_BACKEND_DUMMY

    def __str__(self):
        return ffi.string(libbladeRF.bladerf_backend_str(self.value)).decode()


class DevInfo(collections.namedtuple("DevInfo", [
                "backend", "serial", "usb_bus", "usb_addr", "instance"])):
    @staticmethod
    def from_struct(devinfo):
        return DevInfo(Backend(devinfo.backend),
                       bytes(ffi.string(devinfo.serial)),
                       devinfo.usb_bus,
                       devinfo.usb_addr,
                       devinfo.instance)

    def to_struct(self):
        return ffi.new("struct bladerf_devinfo *", [
                        Backend(self.backend).value, self.serial, self.usb_bus,
                        self.usb_addr, self.instance])

    struct = property(to_struct)

    @property
    def devstr(self):
        d = self._asdict()
        d['serial'] = self.serial_str
        return ("{backend}:device={usb_bus}:{usb_addr} instance={instance}"
                " serial={serial}".format(**d))

    @property
    def serial_str(self):
        return self.serial.decode()

    def __str__(self):
        return ("Device Information\n" +
                "    backend  {}\n".format(self.backend) +
                "    serial   {}\n".format(self.serial_str) +
                "    usb_bus  {}\n".format(self.usb_bus) +
                "    usb_addr {}\n".format(self.usb_addr) +
                "    instance {}".format(self.instance))

    def __repr__(self):
        return '<DevInfo({})>'.format(self.devstr)


class Range(collections.namedtuple("Range", ["min", "max", "step", "scale"])):
    @staticmethod
    def from_struct(_range):
        return Range(_range.min, _range.max, _range.step, _range.scale)

    def __str__(self):
        return ("Range\n" +
                "    min   {}\n".format(self.min) +
                "    max   {}\n".format(self.max) +
                "    step  {}\n".format(self.step) +
                "    scale {}\n".format(self.scale))

    def __repr__(self):
        return '<Range(min={min},max={max},step={step},scale={scale})>'.format(
            **self._asdict())


class DeviceSpeed(enum.Enum):
    Unknown = libbladeRF.BLADERF_DEVICE_SPEED_UNKNOWN
    High = libbladeRF.BLADERF_DEVICE_SPEED_HIGH
    Super = libbladeRF.BLADERF_DEVICE_SPEED_SUPER


class GainMode(enum.Enum):
    Default = libbladeRF.BLADERF_GAIN_DEFAULT
    Manual = libbladeRF.BLADERF_GAIN_MGC
    FastAttack_AGC = libbladeRF.BLADERF_GAIN_FASTATTACK_AGC
    SlowAttack_AGC = libbladeRF.BLADERF_GAIN_SLOWATTACK_AGC
    Hybrid_AGC = libbladeRF.BLADERF_GAIN_HYBRID_AGC

    def __str__(self):
        return self.name

    def __int__(self):
        return self.value


class Direction(enum.Enum):
    TX = libbladeRF.BLADERF_TX
    RX = libbladeRF.BLADERF_RX


class ChannelLayout(enum.Enum):
    RX_X1 = libbladeRF.BLADERF_RX_X1
    TX_X1 = libbladeRF.BLADERF_TX_X1
    RX_X2 = libbladeRF.BLADERF_RX_X2
    TX_X2 = libbladeRF.BLADERF_TX_X2


class Correction(enum.Enum):
    DCOFF_I = libbladeRF.BLADERF_CORR_DCOFF_I
    DCOFF_Q = libbladeRF.BLADERF_CORR_DCOFF_Q
    PHASE = libbladeRF.BLADERF_CORR_PHASE
    GAIN = libbladeRF.BLADERF_CORR_GAIN


class Format(enum.Enum):
    SC16_Q11 = libbladeRF.BLADERF_FORMAT_SC16_Q11
    SC16_Q11_META = libbladeRF.BLADERF_FORMAT_SC16_Q11_META
    PACKET_META = libbladeRF.BLADERF_FORMAT_PACKET_META
    SC8_Q7 = libbladeRF.BLADERF_FORMAT_SC8_Q7
    SC8_Q7_META = libbladeRF.BLADERF_FORMAT_SC8_Q7_META


class Loopback(enum.Enum):
    Disabled = libbladeRF.BLADERF_LB_NONE
    Firmware = libbladeRF.BLADERF_LB_FIRMWARE
    BB_TXLPF_RXVGA2 = libbladeRF.BLADERF_LB_BB_TXLPF_RXVGA2
    BB_TXVGA1_RXVGA2 = libbladeRF.BLADERF_LB_BB_TXVGA1_RXVGA2
    BB_TXLPF_RXLPF = libbladeRF.BLADERF_LB_BB_TXLPF_RXLPF
    BB_TXVGA1_RXLPF = libbladeRF.BLADERF_LB_BB_TXVGA1_RXLPF
    RF_LNA1 = libbladeRF.BLADERF_LB_RF_LNA1
    RF_LNA2 = libbladeRF.BLADERF_LB_RF_LNA2
    RF_LNA3 = libbladeRF.BLADERF_LB_RF_LNA3
    RFIC_BIST = libbladeRF.BLADERF_LB_RFIC_BIST

    def __str__(self):
        return self.name


class RXMux(enum.Enum):
    Invalid = libbladeRF.BLADERF_RX_MUX_INVALID
    Baseband = libbladeRF.BLADERF_RX_MUX_BASEBAND
    Counter_12bit = libbladeRF.BLADERF_RX_MUX_12BIT_COUNTER
    Counter_32bit = libbladeRF.BLADERF_RX_MUX_32BIT_COUNTER
    Digital_Loopback = libbladeRF.BLADERF_RX_MUX_DIGITAL_LOOPBACK

    def __str__(self):
        return self.name


class ClockSelect(enum.Enum):
    Unknown = -99
    VCTCXO = libbladeRF.CLOCK_SELECT_ONBOARD
    External = libbladeRF.CLOCK_SELECT_EXTERNAL

    def __str__(self):
        return self.name


class RSSI(collections.namedtuple("RSSI", ["preamble", "symbol"])):
    def __str__(self):
        return ("RSSI\n" +
                "    preamble {preamble} dB\n" +
                "    symbol   {symbol} dB\n").format(**self._asdict())

    def __repr__(self):
        return ("<RSSI(preamble={preamble}," +
                "symbol={symbol})>").format(**self._asdict())


class PowerSource(enum.Enum):
    Unknown = libbladeRF.BLADERF_UNKNOWN
    DC_Barrel = libbladeRF.BLADERF_PS_DC
    USB_VBUS = libbladeRF.BLADERF_PS_USB_VBUS

    def __str__(self):
        return self.name


class PMICRegister(enum.Enum):
    Configuration = libbladeRF.BLADERF_PMIC_CONFIGURATION
    Voltage_shunt = libbladeRF.BLADERF_PMIC_VOLTAGE_SHUNT
    Voltage_bus = libbladeRF.BLADERF_PMIC_VOLTAGE_BUS
    Power = libbladeRF.BLADERF_PMIC_POWER
    Current = libbladeRF.BLADERF_PMIC_CURRENT
    Calibration = libbladeRF.BLADERF_PMIC_CALIBRATION

    @property
    def ctype(self):
        """C type definition for the pointer to the returned value."""
        if self in [self.Configuration, self.Calibration]:
            return "uint16_t *"
        else:
            return "float *"


class Serial(collections.namedtuple("Serial", ["serial"])):
    @staticmethod
    def from_struct(serial):
        return Serial(ffi.string(serial.serial).decode())

    def __str__(self):
        return self.serial


###############################################################################


class BladeRFError(Exception):
    """A general libbladeRF exception occurred"""
    def __init__(self, code=0, msg=None):
        if isinstance(code, str):
            # SubclassedError("user message")
            msg = code
            code = 0

        if code == 0:
            i_am = self.__class__.__name__
            code = getattr(ReturnCode, i_am).value

        self.errstr = ffi.string(libbladeRF.bladerf_strerror(code)).decode()

        if msg is not None:
            self.errstr += " (" + msg + ")"

    def __str__(self):
        return self.errstr


class UnexpectedError(BladeRFError):
    """An unexpected failure occurred"""


class RangeError(BladeRFError):
    """Provided parameter is out of range"""


class InvalError(BladeRFError):
    """Invalid operation/parameter"""


class MemError(BladeRFError):
    """Memory allocation error"""


class IOError(BladeRFError):
    """File/Device I/O error"""


class TimeoutError(BladeRFError):
    """Operation timed out"""


class NoDevError(BladeRFError):
    """No device(s) available"""


class UnsupportedError(BladeRFError):
    """Operation not supported"""


class MisalignedError(BladeRFError):
    """Misaligned flash access"""


class ChecksumError(BladeRFError):
    """Invalid checksum"""


class NoFileError(BladeRFError):
    """File not found"""


class UpdateFPGAError(BladeRFError):
    """An FPGA update is required"""


class UpdateFWError(BladeRFError):
    """A firmware update is required"""


class TimePastError(BladeRFError):
    """Requested timestamp is in the past"""


class QueueFullError(BladeRFError):
    """Could not enqueue data into full queue"""


class FPGAOpError(BladeRFError):
    """An FPGA operation reported failure"""


class PermissionError(BladeRFError):
    """Insufficient permissions for the requested operation"""


class WouldBlockError(BladeRFError):
    """Operation would block, but has been requested to be non-blocking."""


class NotInitError(BladeRFError):
    """Device insufficiently initialized for operation"""


def _check_error(err):
    """If the result code indicates an error, raise an exception"""
    if err < 0:
        try:
            errorval = ReturnCode(err)
        except ValueError:
            errorval = ReturnCode(0)

        raise globals()[errorval.name](err)


###############################################################################


def set_verbosity(level):
    libbladeRF.bladerf_log_set_verbosity(level)


def version():
    version = ffi.new("struct bladerf_version *")
    libbladeRF.bladerf_version(version)
    return Version.from_struct(version)


def get_device_list():
    devices = ffi.new("struct bladerf_devinfo *[1]")
    ret = libbladeRF.bladerf_get_device_list(devices)
    _check_error(ret)
    devinfos = [DevInfo.from_struct(devices[0][i]) for i in range(ret)]
    libbladeRF.bladerf_free_device_list(devices[0])
    return devinfos


def get_bootloader_list():
    devices = ffi.new("struct bladerf_devinfo *[1]")
    ret = libbladeRF.bladerf_get_bootloader_list(devices)
    _check_error(ret)
    devinfos = [DevInfo.from_struct(devices[0][i]) for i in range(ret)]
    libbladeRF.bladerf_free_device_list(devices[0])
    return devinfos


def load_fw_from_bootloader(device_identifier=None, backend=0, bus=0, addr=0,
                            file=None):
    ret = libbladeRF.bladerf_load_fw_from_bootloader(
        device_identifier.encode(), backend, bus, addr, file.encode())
    _check_error(ret)


###############################################################################


RX = 0x0
TX = 0x1


def CHANNEL_RX(ch):
    return (ch << 1) | RX


def CHANNEL_TX(ch):
    return (ch << 1) | TX


# defaults for PLL on bladeRF 2
BLADERF_VCTCXO_FREQUENCY = 38.4e6
BLADERF_REFIN_DEFAULT = 10.0e6

BLADERF_SERIAL_LENGTH = 33


###############################################################################


class BladeRF:
    """Python class for interacting with bladeRF boards.

    Example usage:

    >>> d = bladerf.BladeRF()
    >>> d.board_name
    'bladerf1'
    """

    def __init__(self, device_identifier=None, devinfo=None):
        self.dev = ffi.new("struct bladerf *[1]")
        self.open(device_identifier, devinfo)

    def __repr__(self):
        return '<BladeRF({!r})>'.format(self.devinfo)

    # Open, close, devinfo

    def open(self, device_identifier=None, devinfo=None):
        if devinfo is not None:
            ret = libbladeRF.bladerf_open_with_devinfo(self.dev,
                                                       devinfo.struct)
        elif device_identifier is not None:
            ret = libbladeRF.bladerf_open(self.dev, device_identifier.encode())
        else:
            ret = libbladeRF.bladerf_open(self.dev, ffi.NULL)
        _check_error(ret)

    def close(self):
        libbladeRF.bladerf_close(self.dev[0])

    def get_devinfo(self):
        devinfo = ffi.new("struct bladerf_devinfo *")
        ret = libbladeRF.bladerf_get_devinfo(self.dev[0], devinfo)
        _check_error(ret)
        return DevInfo.from_struct(devinfo)

    devinfo = property(get_devinfo,
                       doc="Device information as a DevInfo structure")

    # Device properties

    def get_device_speed(self):
        return DeviceSpeed(libbladeRF.bladerf_device_speed(self.dev[0]))

    device_speed = property(get_device_speed,
                            doc="USB speed of the open device")

    def get_serial_struct(self):
        serial = ffi.new("struct bladerf_serial *")
        ret = libbladeRF.bladerf_get_serial_struct(self.dev[0], serial)
        _check_error(ret)
        return Serial.from_struct(serial)

    def get_serial(self):
        sstruct = self.get_serial_struct()
        return sstruct.serial

    serial = property(get_serial, doc="Serial number of the open device")

    def get_fpga_size(self):
        fpga_size = ffi.new("bladerf_fpga_size *")
        ret = libbladeRF.bladerf_get_fpga_size(self.dev[0], fpga_size)
        _check_error(ret)
        return fpga_size[0]

    fpga_size = property(get_fpga_size, doc="FPGA size in kLE")

    def get_flash_size(self):
        flash_size = ffi.new("uint32_t *")
        is_guess = ffi.new("bool *")
        ret = libbladeRF.bladerf_get_flash_size(self.dev[0], flash_size,
                                                is_guess)
        _check_error(ret)
        return (flash_size[0], is_guess[0])

    flash_size = property(get_flash_size, doc="Flash size in bytes")

    def is_fpga_configured(self):
        ret = libbladeRF.bladerf_is_fpga_configured(self.dev[0])
        _check_error(ret)
        return bool(ret)

    fpga_configured = property(is_fpga_configured,
                               doc="True if the device's FPGA is configured")

    def get_fpga_version(self):
        version = ffi.new("struct bladerf_version *")
        ret = libbladeRF.bladerf_fpga_version(self.dev[0], version)
        _check_error(ret)
        return Version.from_struct(version)

    fpga_version = property(get_fpga_version, doc="FPGA version information")

    def get_fw_version(self):
        version = ffi.new("struct bladerf_version *")
        ret = libbladeRF.bladerf_fw_version(self.dev[0], version)
        _check_error(ret)
        return Version.from_struct(version)

    fw_version = property(get_fw_version, doc="Firmware version information")

    def get_board_name(self):
        board_name = libbladeRF.bladerf_get_board_name(self.dev[0])
        return ffi.string(board_name).decode()

    board_name = property(get_board_name,
                          doc="The board model name, as a string")

    def trim_dac_write(self, val):
        libbladeRF.bladerf_trim_dac_write(self.dev[0], val)

    def trim_dac_read(self):
        try:
            trim_val = ffi.new("uint16_t *")
            ret = libbladeRF.bladerf_trim_dac_read(self.dev[0], trim_val)
            _check_error(ret)
            return trim_val[0]
        except UnsupportedError:
            return None

    def get_channel_count(self, direction):
        return libbladeRF.bladerf_get_channel_count(self.dev[0], direction)

    @property
    def rx_channel_count(self):
        """Number of RX channels on the device"""
        return self.get_channel_count(libbladeRF.BLADERF_RX)

    @property
    def tx_channel_count(self):
        """Number of TX channels on the device"""
        return self.get_channel_count(libbladeRF.BLADERF_TX)

    # Enable/Disable

    def enable_module(self, ch, enable):
        ret = libbladeRF.bladerf_enable_module(self.dev[0], ch, bool(enable))
        _check_error(ret)

    # Gain

    def set_gain(self, ch, gain):
        ret = libbladeRF.bladerf_set_gain(self.dev[0], ch, gain)
        _check_error(ret)

    def get_gain(self, ch):
        gain = ffi.new("bladerf_gain *")
        ret = libbladeRF.bladerf_get_gain(self.dev[0], ch, gain)
        _check_error(ret)
        return gain[0]

    def set_gain_mode(self, ch, mode):
        ret = libbladeRF.bladerf_set_gain_mode(self.dev[0], ch, int(mode))
        _check_error(ret)

    def get_gain_mode(self, ch):
        try:
            mode = ffi.new("bladerf_gain_mode *")
            ret = libbladeRF.bladerf_get_gain_mode(self.dev[0], ch, mode)
            _check_error(ret)
            return GainMode(mode[0])
        except UnsupportedError:
            return None

    def get_gain_modes(self, ch):
        modes_arr = ffi.new("struct bladerf_gain_modes **")
        ret = libbladeRF.bladerf_get_gain_modes(self.dev[0], ch, modes_arr)
        _check_error(ret)
        return [GainMode(modes_arr[0][i].mode) for i in range(ret)]

    def get_gain_range(self, ch):
        _range_ptr = ffi.new("struct bladerf_range **")
        ret = libbladeRF.bladerf_get_gain_range(self.dev[0], ch, _range_ptr)
        _check_error(ret)
        return Range.from_struct(_range_ptr[0])

    def set_gain_stage(self, ch, stage, gain):
        ret = libbladeRF.bladerf_set_gain_stage(self.dev[0], ch,
                                                stage.encode(), gain)
        _check_error(ret)

    def get_gain_stage(self, ch, stage):
        gain = ffi.new("bladerf_gain *")
        ret = libbladeRF.bladerf_get_gain_stage(self.dev[0], ch,
                                                stage.encode(), gain)
        _check_error(ret)
        return gain[0]

    def get_gain_stage_range(self, ch, stage):
        _range_ptr = ffi.new("struct bladerf_range **")
        ret = libbladeRF.bladerf_get_gain_stage_range(self.dev[0], ch,
                                                      stage.encode(),
                                                      _range_ptr)
        _check_error(ret)
        return Range.from_struct(_range_ptr[0])

    def get_gain_stages(self, ch):
        ret = libbladeRF.bladerf_get_gain_stages(self.dev[0], ch, ffi.NULL, 0)
        _check_error(ret)
        stages_arr = ffi.new("const char *[]", ret)
        ret = libbladeRF.bladerf_get_gain_stages(self.dev[0], ch, stages_arr,
                                                 ret)
        _check_error(ret)
        return [ffi.string(stages_arr[i]).decode() for i in range(ret)]

    # Sample rate

    def set_sample_rate(self, ch, rate):
        actual_rate = ffi.new("bladerf_sample_rate *")
        ret = libbladeRF.bladerf_set_sample_rate(self.dev[0], ch, int(rate),
                                                 actual_rate)
        _check_error(ret)
        return actual_rate[0]

    def set_rational_sample_rate(self, ch, rational_rate):
        raise NotImplementedError()

    def get_sample_rate(self, ch):
        rate = ffi.new("bladerf_sample_rate *")
        ret = libbladeRF.bladerf_get_sample_rate(self.dev[0], ch, rate)
        _check_error(ret)
        return rate[0]

    def get_rational_sample_rate(self, ch):
        raise NotImplementedError()

    def get_sample_rate_range(self, ch):
        _range_ptr = ffi.new("struct bladerf_range **")
        ret = libbladeRF.bladerf_get_sample_rate_range(self.dev[0], ch,
                                                       _range_ptr)
        _check_error(ret)
        return Range.from_struct(_range_ptr[0])

    # Bandwidth

    def set_bandwidth(self, ch, bandwidth):
        actual_bandwidth = ffi.new("bladerf_bandwidth *")
        ret = libbladeRF.bladerf_set_bandwidth(self.dev[0], ch, int(bandwidth),
                                               actual_bandwidth)
        _check_error(ret)
        return actual_bandwidth[0]

    def get_bandwidth(self, ch):
        bandwidth = ffi.new("bladerf_bandwidth *")
        ret = libbladeRF.bladerf_get_bandwidth(self.dev[0], ch, bandwidth)
        _check_error(ret)
        return bandwidth[0]

    def get_bandwidth_range(self, ch):
        _range_ptr = ffi.new("struct bladerf_range **")
        ret = libbladeRF.bladerf_get_bandwidth_range(self.dev[0], ch,
                                                     _range_ptr)
        _check_error(ret)
        return Range.from_struct(_range_ptr[0])

    # Frequency

    def set_frequency(self, ch, frequency):
        ret = libbladeRF.bladerf_set_frequency(self.dev[0], ch, int(frequency))
        _check_error(ret)

    def get_frequency(self, ch):
        frequency = ffi.new("bladerf_frequency *")
        ret = libbladeRF.bladerf_get_frequency(self.dev[0], ch, frequency)
        _check_error(ret)
        return frequency[0]

    def select_band(self, ch, frequency):
        ret = libbladeRF.bladerf_select_band(self.dev[0], ch, int(frequency))
        _check_error(ret)

    def get_frequency_range(self, ch):
        _range_ptr = ffi.new("struct bladerf_range **")
        ret = libbladeRF.bladerf_get_frequency_range(self.dev[0], ch,
                                                     _range_ptr)
        _check_error(ret)
        return Range.from_struct(_range_ptr[0])

    # RF Ports

    def set_rf_port(self, ch, port):
        ret = libbladeRF.bladerf_set_rf_port(self.dev[0], ch, port.encode())
        _check_error(ret)

    def get_rf_port(self, ch):
        try:
            port = ffi.new("const char *[1]")
            ret = libbladeRF.bladerf_get_rf_port(self.dev[0], ch, port)
            _check_error(ret)
            return ffi.string(port[0]).decode()
        except UnsupportedError:
            return None

    def get_rf_ports(self, ch):
        try:
            count = libbladeRF.bladerf_get_rf_ports(self.dev[0], ch, ffi.NULL,
                                                    0)
            _check_error(count)
            ports_arr = ffi.new("const char *[]", count)
            ret = libbladeRF.bladerf_get_rf_ports(self.dev[0], ch, ports_arr,
                                                  count)
            _check_error(ret)
            return [ffi.string(ports_arr[i]).decode() for i in range(ret)]
        except UnsupportedError:
            return []

    # Sample Loopback

    def get_loopback_modes(self):
        ret = libbladeRF.bladerf_get_loopback_modes(self.dev[0], ffi.NULL)
        _check_error(ret)
        modes_arr = ffi.new("struct bladerf_loopback_modes *[]", ret)
        ret = libbladeRF.bladerf_get_loopback_modes(self.dev[0], modes_arr)
        _check_error(ret)
        return [Loopback(modes_arr[0][i].mode) for i in range(ret)]

    loopback_modes = property(get_loopback_modes,
                              doc="Supported loopback modes")

    def is_loopback_mode_supported(self, lb):
        if isinstance(lb, Loopback):
            lb = lb.value
        ret = libbladeRF.bladerf_is_loopback_mode_supported(self.dev[0], lb)
        _check_error(ret)
        return bool(ret)

    def get_loopback(self):
        lb = ffi.new("bladerf_loopback *")
        ret = libbladeRF.bladerf_get_loopback(self.dev[0], lb)
        _check_error(ret)
        return Loopback(lb[0])

    def set_loopback(self, lb):
        if isinstance(lb, Loopback):
            lb = lb.value
        ret = libbladeRF.bladerf_set_loopback(self.dev[0], lb)
        _check_error(ret)

    loopback = property(get_loopback, set_loopback, doc="Loopback selection")

    # Trigger TBD

    # Sample RX Mux

    def get_rx_mux(self):
        mux = ffi.new("bladerf_rx_mux *")
        ret = libbladeRF.bladerf_get_rx_mux(self.dev[0], mux)
        _check_error(ret)
        return RXMux(mux[0])

    def set_rx_mux(self, rx_mux):
        if isinstance(rx_mux, RXMux):
            rx_mux = rx_mux.value
        ret = libbladeRF.bladerf_set_rx_mux(self.dev[0], rx_mux)
        _check_error(ret)

    rx_mux = property(get_rx_mux, set_rx_mux, doc="RX Multiplexer selection")

    # DC/Phase/Gain Correction

    def get_correction(self, ch, corr):
        value = ffi.new("bladerf_correction_value *")
        ret = libbladeRF.bladerf_get_correction(self.dev[0], ch, corr.value,
                                                value)
        _check_error(ret)
        return value[0]

    def set_correction(self, ch, corr, value):
        ret = libbladeRF.bladerf_set_correction(self.dev[0], ch, corr.value,
                                                value)
        _check_error(ret)

    # Streaming format

    def interleave_stream_buffer(self, layout, format, buffer_size, samples):
        raise NotImplementedError()

    def deinterleave_stream_buffer(self, layout, format, buffer_size, samples):
        raise NotImplementedError()

    # Streaming

    def sync_config(self, layout, fmt, num_buffers, buffer_size, num_transfers,
                    stream_timeout):
        if fmt not in [Format.SC16_Q11, Format.SC8_Q7]:
            raise NotImplementedError("Format not supported by binding.")

        ret = libbladeRF.bladerf_sync_config(self.dev[0],
                                             layout.value,
                                             fmt.value,
                                             num_buffers,
                                             buffer_size,
                                             num_transfers,
                                             stream_timeout)
        _check_error(ret)

    def sync_tx(self, buf, num_samples, timeout_ms=None):
        ret = libbladeRF.bladerf_sync_tx(self.dev[0],
                                         ffi.from_buffer(buf),
                                         num_samples,
                                         ffi.NULL,
                                         timeout_ms or 0)
        _check_error(ret)

    def sync_rx(self, buf, num_samples, timeout_ms=None):
        ret = libbladeRF.bladerf_sync_rx(self.dev[0],
                                         ffi.from_buffer(buf),
                                         num_samples,
                                         ffi.NULL,
                                         timeout_ms or 0)
        _check_error(ret)

    # FPGA/Firmware Loading/Flashing

    def load_fpga(self, image_path):
        ret = libbladeRF.bladerf_load_fpga(self.dev[0], image_path.encode())
        _check_error(ret)

    def flash_fpga(self, image_path):
        ret = libbladeRF.bladerf_flash_fpga(self.dev[0], image_path.encode())
        _check_error(ret)

    def erase_stored_fpga(self):
        ret = libbladeRF.bladerf_erase_stored_fpga(self.dev[0])
        _check_error(ret)

    def flash_firmware(self, image_path):
        ret = libbladeRF.bladerf_flash_firmware(self.dev[0],
                                                image_path.encode())
        _check_error(ret)

    def device_reset(self):
        ret = libbladeRF.bladerf_device_reset(self.dev[0])
        _check_error(ret)

    # VCTCXO Tamer Mode

    def set_vctcxo_tamer_mode(self, mode):
        raise NotImplementedError()

    def get_vctcxo_tamer_mode(self):
        raise NotImplementedError()

    vctcxo_tamer_mode = property(get_vctcxo_tamer_mode, set_vctcxo_tamer_mode)

    def get_vctcxo_trim(self):
        raise NotImplementedError()

    vctcxo = property(get_vctcxo_trim)

    # Tuning Mode

    def set_tuning_mode(self, mode):
        raise NotImplementedError()

    def get_tuning_mode(self):
        raise NotImplementedError()

    tuning_mode = property(get_tuning_mode, set_tuning_mode)

    # Bias tee

    def get_bias_tee(self, ch):
        try:
            state = ffi.new("bool *")
            ret = libbladeRF.bladerf_get_bias_tee(self.dev[0], ch, state)
            _check_error(ret)
            return bool(state[0])
        except UnsupportedError:
            return None

    def set_bias_tee(self, ch, enable):
        ret = libbladeRF.bladerf_set_bias_tee(self.dev[0], ch, enable)
        _check_error(ret)

    # RFIC

    def get_rfic_temperature(self):
        try:
            val = ffi.new("float *")
            ret = libbladeRF.bladerf_get_rfic_temperature(self.dev[0], val)
            _check_error(ret)
            return val[0]
        except UnsupportedError:
            return None

    rfic_temperature = property(get_rfic_temperature,
                                doc="RFIC internal temperature")

    def get_rfic_rssi(self, ch):
        try:
            preamble = ffi.new("int32_t *")
            symbol = ffi.new("int32_t *")
            ret = libbladeRF.bladerf_get_rfic_rssi(self.dev[0], ch,
                                                   preamble, symbol)
            _check_error(ret)
            return RSSI(preamble[0], symbol[0])
        except UnsupportedError:
            return None

    def get_rfic_ctrl_out(self):
        try:
            ctrl_out = ffi.new("uint8_t *")
            ret = libbladeRF.bladerf_get_rfic_ctrl_out(self.dev[0], ctrl_out)
            _check_error(ret)
            return ctrl_out[0]
        except UnsupportedError:
            return None

    rfic_ctrl_out = property(get_rfic_ctrl_out,
                             doc="RFIC CTRL_OUT status pins")

    # Phase Detector/Frequency Synthesizer

    def get_pll_lock_state(self):
        try:
            locked = ffi.new("bool *")
            ret = libbladeRF.bladerf_get_pll_lock_state(self.dev[0], locked)
            _check_error(ret)
            return bool(locked[0])
        except UnsupportedError:
            return None

    pll_locked = property(get_pll_lock_state,
                          doc="PLL lock indication")

    def get_pll_enable(self):
        try:
            enable = ffi.new("bool *")
            ret = libbladeRF.bladerf_get_pll_enable(self.dev[0], enable)
            _check_error(ret)
            return bool(enable[0])
        except UnsupportedError:
            return None

    def set_pll_enable(self, enable):
        ret = libbladeRF.bladerf_set_pll_enable(self.dev[0], enable)
        _check_error(ret)

    pll_enable = property(get_pll_enable, set_pll_enable,
                          doc="PLL Enable")

    def get_pll_refclk(self):
        try:
            freq = ffi.new("uint64_t *")
            ret = libbladeRF.bladerf_get_pll_refclk(self.dev[0], freq)
            _check_error(ret)
            return int(freq[0])
        except UnsupportedError:
            return None

    def set_pll_refclk(self, freq):
        ret = libbladeRF.bladerf_set_pll_refclk(self.dev[0], freq)
        _check_error(ret)

    pll_refclk = property(get_pll_refclk, set_pll_refclk,
                          doc="PLL Reference Clock frequency")

    # Power source

    def get_power_source(self):
        try:
            val = ffi.new("bladerf_power_sources *")
            ret = libbladeRF.bladerf_get_power_source(self.dev[0], val)
            _check_error(ret)
            return PowerSource(val[0])
        except UnsupportedError:
            return PowerSource.Unknown

    power_source = property(get_power_source,
                            doc="Active power source for device")

    # Clock select

    def get_clock_select(self):
        try:
            sel = ffi.new("bladerf_clock_select *")
            ret = libbladeRF.bladerf_get_clock_select(self.dev[0], sel)
            _check_error(ret)
            return ClockSelect(sel[0])
        except UnsupportedError:
            return ClockSelect.Unknown

    def set_clock_select(self, sel):
        if isinstance(sel, ClockSelect):
            sel = sel.value
        ret = libbladeRF.bladerf_set_clock_select(self.dev[0], sel)
        _check_error(ret)

    clock_select = property(get_clock_select, set_clock_select,
                            doc="System clock input selection")

    # Clock output

    def get_clock_output(self):
        try:
            state = ffi.new("bool *")
            ret = libbladeRF.bladerf_get_clock_output(self.dev[0], state)
            _check_error(ret)
            return bool(state[0])
        except UnsupportedError:
            return None

    def set_clock_output(self, enable):
        ret = libbladeRF.bladerf_set_clock_output(self.dev[0], enable)
        _check_error(ret)

    clock_output = property(get_clock_output, set_clock_output,
                            doc="Clock output enable")

    # Current/power monitor IC

    def get_pmic_register(self, reg):
        try:
            if not isinstance(reg, PMICRegister):
                reg = PMICRegister(reg)
            val = ffi.new(reg.ctype)
            ret = libbladeRF.bladerf_get_pmic_register(self.dev[0], reg.value,
                                                       val)
            _check_error(ret)
            return val[0]
        except UnsupportedError:
            return None

    @property
    def bus_voltage(self):
        """Main power bus voltage, in Volts"""
        return self.get_pmic_register(PMICRegister.Voltage_bus)

    @property
    def bus_current(self):
        """Main power bus current, in Amps"""
        return self.get_pmic_register(PMICRegister.Current)

    @property
    def bus_power(self):
        """Main power bus power, in Watts"""
        return self.get_pmic_register(PMICRegister.Power)

    # Convenience objects

    class _Channel:
        """Container for bladeRF channel properties"""

        def __init__(self, dev, channel):
            self.dev = dev
            self.channel = channel

            if (channel >> 1) not in range(dev.get_channel_count(self.is_tx)):
                raise UnsupportedError("Invalid channel: " + str(channel))

        def __str__(self):
            return "Channel {}{}".format(
                "TX" if self.is_tx else "RX", (self.channel >> 1) + 1)

        def __repr__(self):
            return '<Channel({},CHANNEL_{}({}))>'.format(
                repr(self.dev), "TX" if self.is_tx else "RX",
                self.channel >> 1)

        @property
        def is_tx(self):
            """Returns True if this is a TX channel, False otherwise"""
            return bool(self.channel & TX)

        @property
        def gain(self):
            """Gain, in dB"""
            return self.dev.get_gain(self.channel)

        @gain.setter
        def gain(self, value):
            return self.dev.set_gain(self.channel, value)

        @property
        def gain_mode(self):
            """Gain mode"""
            return self.dev.get_gain_mode(self.channel)

        @gain_mode.setter
        def gain_mode(self, value):
            return self.dev.set_gain_mode(self.channel, value)

        @property
        def gain_modes(self):
            """List of supported gain modes"""
            return self.dev.get_gain_modes(self.channel)

        @property
        def rssi(self):
            """RSSI report from RFIC"""
            return self.dev.get_rfic_rssi(self.channel)

        @property
        def preamble_rssi(self):
            """Preamble RSSI reading (latched at algorithm reset)"""
            if self.rssi is not None:
                return self.rssi.preamble
            else:
                return None

        @property
        def symbol_rssi(self):
            """Symbol RSSI reading (most recent value)"""
            if self.rssi is not None:
                return self.rssi.symbol
            else:
                return None

        @property
        def sample_rate(self):
            """Sample rate, in samples/second"""
            return self.dev.get_sample_rate(self.channel)

        @sample_rate.setter
        def sample_rate(self, value):
            return self.dev.set_sample_rate(self.channel, value)

        @property
        def sample_rate_range(self):
            """Range of sample rates supported by this channel"""
            return self.dev.get_sample_rate_range(self.channel)

        @property
        def bandwidth(self):
            """Channel bandwidth in Hz"""
            return self.dev.get_bandwidth(self.channel)

        @bandwidth.setter
        def bandwidth(self, value):
            return self.dev.set_bandwidth(self.channel, value)

        @property
        def bandwidth_range(self):
            """Range of bandwidths supported by this channel"""
            return self.dev.get_bandwidth_range(self.channel)

        @property
        def frequency(self):
            """Frequency in Hz"""
            return self.dev.get_frequency(self.channel)

        @frequency.setter
        def frequency(self, value):
            return self.dev.set_frequency(self.channel, value)

        @property
        def frequency_range(self):
            """Range of frequencies supported by this channel"""
            return self.dev.get_frequency_range(self.channel)

        @property
        def rf_port(self):
            """Active RF port"""
            return self.dev.get_rf_port(self.channel)

        @rf_port.setter
        def rf_port(self, value):
            return self.dev.set_rf_port(self.channel, value)

        @property
        def rf_ports(self):
            """List of available RF ports"""
            return self.dev.get_rf_ports(self.channel)

        @property
        def bias_tee(self):
            """Bias-T control"""
            return self.dev.get_bias_tee(self.channel)

        @bias_tee.setter
        def bias_tee(self, value):
            return self.dev.set_bias_tee(self.channel, value)

        def set_enable(self, value):
            return self.dev.enable_module(self.channel, value)

        enable = property(fset=set_enable, doc="Enable/disable RF chain")

    def Channel(self, ch):
        """Returns an object for a given subchannel.

        Usage:
        >>> d = bladerf.BladeRF()
        >>> rx = d.Channel(bladerf.CHANNEL_RX(0))

        See help(bladerf.BladeRF._Channel) or help(rx) for more information.
        """
        return self._Channel(self, ch)



#!/usr/bin/env python3
#
# Convert a given bladeRF CSV samples file to SC16Q11 binary format.
#

import csv
import struct
import sys
from   pathlib import Path

def csv2bin( csvfile = None, binfile = None ):
    with open(csvfile, 'r') as c:
        with open(binfile, 'wb') as b:
            csvreader = csv.reader(c, delimiter=',')
            count = 0
            for row in csvreader:
                for col in row:
                    # Little endian 16-bit integers (shorts)
                    b.write( struct.pack("<h", int(col.strip())) )
                count += 1
            print( "Processed", str(count), "samples." )

if( len(sys.argv) != 3 ):
    print( "Usage:", sys.argv[0], "in.csv out.bin" )
    sys.exit(-1)
else:
    csvfile = sys.argv[1]
    binfile = sys.argv[2]
    if( Path(csvfile).is_file() ):
        csv2bin( csvfile, binfile )
    else:
        print( "File does not exist:", csvfile )
        sys.exit(-1)



#!/usr/bin/env python3
###############################################################################
#
# Copyright (c) 2018-present Nuand LLC
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################
#
# Basic example of using bladeRF Python bindings for full duplex TX/RX.
# Review the companion txrx.ini to adjust configuration options.
#
###############################################################################

import sys
import os
import threading

from multiprocessing.pool import ThreadPool
from configparser         import ConfigParser

from bladerf              import _bladerf

# =============================================================================
# Close the device and exit
# =============================================================================
def shutdown( error = 0, board = None ):
    print( "Shutting down with error code: " + str(error) )
    if( board != None ):
        board.close()
    sys.exit(error)


# =============================================================================
# Version information
# =============================================================================
def print_versions( device = None ):
    print( "libbladeRF version: " + str(_bladerf.version()) )
    if( device != None ):
        try:
            print( "Firmware version: " + str(device.get_fw_version()) )
        except:
            print( "Firmware version: " + "ERROR" )
            raise

        try:
            print( "FPGA version: "     + str(device.get_fpga_version()) )
        except:
            print( "FPGA version: "     + "ERROR" )
            raise

    return 0


# =============================================================================
# Search for a bladeRF device attached to the host system
# Returns a bladeRF device handle.
# =============================================================================
def probe_bladerf():
    device = None
    print( "Searching for bladeRF devices..." )
    try:
        devinfos = _bladerf.get_device_list()
        if( len(devinfos) == 1 ):
            device = "{backend}:device={usb_bus}:{usb_addr}".format(**devinfos[0]._asdict())
            print( "Found bladeRF device: " + str(device) )
        if( len(devinfos) > 1 ):
            print( "Unsupported feature: more than one bladeRFs detected." )
            print( "\n".join([str(devinfo) for devinfo in devinfos]) )
            shutdown( error = -1, board = None )
    except _bladerf.BladeRFError:
        print( "No bladeRF devices found." )
        pass

    return device


# =============================================================================
# Load FPGA
# =============================================================================
def load_fpga( device, image ):

    image = os.path.abspath( image )

    if( not os.path.exists(image) ):
        print( "FPGA image does not exist: " + str(image) )
        return -1

    try:
        print( "Loading FPGA image: " + str(image ) )
        device.load_fpga( image )
        fpga_loaded  = device.is_fpga_configured()
        fpga_version = device.get_fpga_version()

        if( fpga_loaded ):
            print( "FPGA successfully loaded. Version: " + str(fpga_version) )

    except _bladerf.BladeRFError:
        print( "Error loading FPGA." )
        raise

    return 0


# =============================================================================
# TRANSMIT
# =============================================================================
def transmit( device, channel : int, freq : int, rate : int, gain : int,
              tx_start = None, rx_done = None,
              txfile : str = '', repeat : int = 1,  ):

    if( device == None ):
        print( "TX: Invalid device handle." )
        return -1

    if( channel == None ):
        print( "TX: Invalid channel." )
        return -1

    if( (rx_done == None) and (repeat < 1) ):
        print( "TX: Configuration settings indicate transmitting indefinitely?" )
        return -1

    if( tx_start != None ):
        print( "TX: waiting until receive thread is ready..." )
        if( not tx_start.wait(60.0) ):
            print( "TX: Timeout occurred while waiting for receiver to " +
                   "become ready." )
            return -1

    # Configure bladeRF
    ch             = device.Channel(channel)
    ch.frequency   = freq
    ch.sample_rate = rate
    ch.gain        = gain

    # Setup stream
    device.sync_config(layout=_bladerf.ChannelLayout.TX_X1,
                       fmt=_bladerf.Format.SC16_Q11,
                       num_buffers=16,
                       buffer_size=8192,
                       num_transfers=8,
                       stream_timeout=3500)

    # Enable module
    print( "TX: Start" )
    ch.enable = True

    # Create buffer
    bytes_per_sample = 4
    buf = bytearray(1024*bytes_per_sample)

    with open(txfile, 'rb') as infile:
        # Read samples from file into buf
        num = infile.readinto(buf)

        # Convert number of bytes read to samples
        num //= bytes_per_sample
        if( num > 0 ):
            repeats_remaining = repeat - 1
            repeat_inf        = (repeat < 1)
            while True:
                # Write to bladeRF
                device.sync_tx(buf, num)

                if( (rx_done != None) and rx_done.is_set() ):
                    break

                if( not repeat_inf ):
                    if( repeats_remaining > 0 ):
                        repeats_remaining -= 1
                    else:
                        break

    # Disable module
    print( "TX: Stop" )
    ch.enable = False

    return 0


# =============================================================================
# RECEIVE
# =============================================================================
def receive(device, channel : int, freq : int, rate : int, gain : int,
            tx_start = None, rx_done = None,
            rxfile : str = '', num_samples : int = 1024):

    status = 0

    if( device == None ):
        print( "RX: Invalid device handle." )
        return -1

    if( channel == None ):
        print( "RX: Invalid channel." )
        return -1

    # Configure BladeRF
    ch             = device.Channel(channel)
    ch.frequency   = freq
    ch.sample_rate = rate
    ch.gain        = gain

    # Setup synchronous stream
    device.sync_config(layout         = _bladerf.ChannelLayout.RX_X1,
                       fmt            = _bladerf.Format.SC16_Q11,
                       num_buffers    = 16,
                       buffer_size    = 8192,
                       num_transfers  = 8,
                       stream_timeout = 3500)

    # Enable module
    print( "RX: Start" )
    ch.enable = True

    # Create receive buffer
    bytes_per_sample = 4
    buf = bytearray(1024*bytes_per_sample)
    num_samples_read = 0

    # Tell TX thread to begin
    if( tx_start != None ):
        tx_start.set()

    # Save the samples
    with open(rxfile, 'wb') as outfile:
        while True:
            if num_samples > 0 and num_samples_read == num_samples:
                break
            elif num_samples > 0:
                num = min(len(buf)//bytes_per_sample,
                          num_samples-num_samples_read)
            else:
                num = len(buf)//bytes_per_sample

            # Read into buffer
            device.sync_rx(buf, num)
            num_samples_read += num

            # Write to file
            outfile.write(buf[:num*bytes_per_sample])

    # Disable module
    print( "RX: Stop" )
    ch.enable = False

    if( rx_done != None ):
        rx_done.set()

    print( "RX: Done" )

    return 0


# =============================================================================
# Load Configuration
# =============================================================================

config = ConfigParser()
config.read('txrx.ini')

# Set libbladeRF verbosity level
verbosity = config.get('common', 'libbladerf_verbosity').upper()
if  ( verbosity == "VERBOSE" ):  _bladerf.set_verbosity( 0 )
elif( verbosity == "DEBUG" ):    _bladerf.set_verbosity( 1 )
elif( verbosity == "INFO" ):     _bladerf.set_verbosity( 2 )
elif( verbosity == "WARNING" ):  _bladerf.set_verbosity( 3 )
elif( verbosity == "ERROR" ):    _bladerf.set_verbosity( 4 )
elif( verbosity == "CRITICAL" ): _bladerf.set_verbosity( 5 )
elif( verbosity == "SILENT" ):   _bladerf.set_verbosity( 6 )
else:
    print( "Invalid libbladerf_verbosity specified in configuration file:",
           verbosity )
    shutdown( error = -1, board = None )


# =============================================================================
# BEGIN !!!
# =============================================================================

uut = probe_bladerf()

if( uut == None ):
    print( "No bladeRFs detected. Exiting." )
    shutdown( error = -1, board = None )

b          = _bladerf.BladeRF( uut )
board_name = b.board_name
fpga_size  = b.fpga_size


if( config.getboolean(board_name + '-load-fpga', 'enable') ):
    print( "Loading FPGA..." )
    try:
        status = load_fpga( b, config.get(board_name + '-load-fpga',
                                          'image_' + str(fpga_size) + 'kle' ) )
    except:
        print( "ERROR loading FPGA." )
        raise

    if( status < 0 ):
        print( "ERROR loading FPGA." )
        shutdown( error = status, board = b )
else:
    print( "Skipping FPGA load due to configuration setting." )

status = print_versions( device = b )

# TODO: can we have >1 rx/tx pool workers because 2x2 MIMO?
rx_pool = ThreadPool(processes=1)
tx_pool = ThreadPool(processes=1)

for s in [ss for ss in config.sections() if board_name + '-' in ss]:

    if( s == board_name + "-load-fpga" ):
        # Don't re-loading the FPGA!
        continue

    # Print the section name
    print( "{:<35s} : ".format(s), end='' )

    if( config.getboolean(s, 'enable') ):

        print( "RUNNING" )

        if( s == board_name + '-tx' ):

            tx_ch   = _bladerf.CHANNEL_TX(config.getint(s, 'tx_channel'))
            tx_freq = int(config.getfloat(s, 'tx_frequency'))
            tx_rate = int(config.getfloat(s, 'tx_samplerate'))
            tx_gain = int(config.getfloat(s, 'tx_gain'))
            tx_rpt  = int(config.getfloat(s, 'tx_repeats'))
            tx_file = config.get(s, 'tx_file')

            # Make this blocking for now ...
            status = tx_pool.apply_async(transmit,
                                         (),
                                         { 'device'   : b,
                                           'channel'  : tx_ch,
                                           'freq'     : tx_freq,
                                           'rate'     : tx_rate,
                                           'gain'     : tx_gain,
                                           'tx_start' : None,
                                           'rx_done'  : None,
                                           'txfile'   : tx_file,
                                           'repeat'   : tx_rpt
                                         }).get()
            if( status < 0 ):
                print( "Transmit operation failed with error " + str(status) )

        elif( s == board_name + '-rx' ):

            rx_ch   = _bladerf.CHANNEL_RX(config.getint(s, 'rx_channel'))
            rx_freq = int(config.getfloat(s, 'rx_frequency'))
            rx_rate = int(config.getfloat(s, 'rx_samplerate'))
            rx_gain = int(config.getfloat(s, 'rx_gain'))
            rx_ns   = int(config.getfloat(s, 'rx_num_samples'))
            rx_file = config.get(s, 'rx_file')

            # Make this blocking for now ...
            status = rx_pool.apply_async(receive,
                                         (),
                                         { 'device'        : b,
                                           'channel'       : rx_ch,
                                           'freq'          : rx_freq,
                                           'rate'          : rx_rate,
                                           'gain'          : rx_gain,
                                           'tx_start'      : None,
                                           'rx_done'       : None,
                                           'rxfile'        : rx_file,
                                           'num_samples'   : rx_ns
                                         }).get()
            if( status < 0 ):
                print( "Receive operation failed with error " + str(status) )

        elif( s == board_name + '-txrx' ):

            rx_channels  = [x.strip() for x in config.get(s, 'rx_channel').split(',')]
            tx_channels  = [x.strip() for x in config.get(s, 'tx_channel').split(',')]
            rx_freqs     = [x.strip() for x in config.get(s, 'rx_frequency').split(',')]
            tx_freqs     = [x.strip() for x in config.get(s, 'tx_frequency').split(',')]

            if( len(rx_channels) != len(tx_channels) ):
                print( "Configuration error in section " + s + ": "
                       "rx_channels and tx_channels must be the same length." )
                shutdown( error = -1, board = b )

            if( len(rx_freqs) != len(tx_freqs) ):
                print( "Configuration error in section " + s + ": "
                       "rx_frequency and tx_frequency must be the same length." )
                shutdown( error = -1, board = b )

            # Create Events for signaling between RX/TX threads
            rx_done  = threading.Event()
            tx_start = threading.Event()

            for ch in range(0, len(rx_channels), 1):
                for freq in range(0, len(rx_freqs), 1):
                    rx_ch   = _bladerf.CHANNEL_RX(int(rx_channels[ch]))
                    rx_freq = int(float(rx_freqs[freq]))
                    rx_rate = int(config.getfloat(s, 'rx_samplerate'))
                    rx_gain = int(config.getfloat(s, 'rx_gain'))
                    rx_ns   = int(config.getfloat(s, 'rx_num_samples'))
                    rx_file = config.get(s, 'rx_file')
                    if( rx_file == "auto" ):
                        rx_file = "rx_" + \
                            "r" + rx_channels[ch] + \
                            "t" + tx_channels[ch] + "_" + \
                            str(int(float(rx_freqs[freq])/(1e6))) + "M.bin"

                    tx_ch   = _bladerf.CHANNEL_TX(int(tx_channels[ch]))
                    tx_freq = int(float(tx_freqs[freq]))
                    tx_rate = int(config.getfloat(s, 'tx_samplerate'))
                    tx_gain = int(config.getfloat(s, 'tx_gain'))
                    tx_rpt  = int(config.getfloat(s, 'tx_repeats'))
                    tx_file = config.get(s, 'tx_file')

                    print( "rx_ch = {:2d} ".format(int(rx_channels[ch])) +
                           "tx_ch = {:2d} ".format(int(tx_channels[ch])) +
                           "rx_freq = {:10d} ".format(rx_freq) +
                           "tx_freq = {:10d} ".format(tx_freq) +
                           "rx_file = " + rx_file )

                    # Start receiver thread
                    rx_result = rx_pool.apply_async(receive,
                                                    (),
                                                    { 'device'      : b,
                                                      'channel'     : rx_ch,
                                                      'freq'        : rx_freq,
                                                      'rate'        : rx_rate,
                                                      'gain'        : rx_gain,
                                                      'tx_start'    : tx_start,
                                                      'rx_done'     : rx_done,
                                                      'rxfile'      : rx_file,
                                                      'num_samples' : rx_ns
                                                    })

                    # Start transmitter thread
                    tx_result = tx_pool.apply_async(transmit,
                                                    (),
                                                    { 'device'   : b,
                                                      'channel'  : tx_ch,
                                                      'freq'     : tx_freq,
                                                      'rate'     : tx_rate,
                                                      'gain'     : tx_gain,
                                                      'tx_start' : tx_start,
                                                      'rx_done'  : rx_done,
                                                      'txfile'   : tx_file,
                                                      'repeat'   : tx_rpt
                                                    })

                    # Wait for RX thread to finish
                    rx_result.wait()
                    status = rx_result.get()

                    if( status < 0 ):
                        print( "Receive operation failed with error " +
                               str(status) )

                    # Wait for TX thread to finish
                    tx_result.wait()
                    status = tx_result.get()

                    if( status < 0 ):
                        print( "Transmit operation failed with error " +
                               str(status) )

                    tx_start.clear()
                    rx_done.clear()

    else:
        print( "SKIPPED [ Disabled ]" )


b.close()
print( "Done!" )



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='bladerf',
    version='1.3.0',
    description='CFFI-based Python 3 binding to libbladeRF',
    long_description=long_description,
    url='https://github.com/Nuand/bladeRF',
    author='Nuand LLC',
    author_email='bladerf@nuand.com',
    license='MIT',
    python_requires='>=3',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Telecommunications Industry',
        'Topic :: Communications :: Ham Radio',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    keywords='bladerf sdr cffi radio libbladerf',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=['cffi'],
    entry_points={
        'console_scripts': [
            'bladerf-tool=bladerf._tool:main',
        ],
    },
    project_urls={
        'Bug Reports': 'https://github.com/Nuand/bladeRF/issues',
        'Source': 'https://github.com/Nuand/bladeRF',
        'Nuand': 'https://nuand.com',
    },
)
